// by Alexander Ilyin
// arguments of the command line

// *log file must not be used here, cause args initializes placement of log file

#include "args.h"
#include <string.h>
#include <stdio.h>
#include "error.h"
#include "console.h"
#include "reg_drag.h"
#include "trayicon.h"
#include "displaying.h"
#include "sorting.h"
#include "string.h"
#include "log_file.h"



#define ARGS_MAX 100
#define ARG_MAX_LENGHT 100


static int args_num = 0;
static char Args[ARGS_MAX][ARG_MAX_LENGHT];

static char* main_arg = "Drag_Watcher.exe";

/*
#define ARG_NONE 0
#define ARG_INT 1
#define ARG_DOUBLE 2
#define ARG_DWORD 3
#define ARG_ZERO_INT 4
#define ARG_UINT 5
*/
enum
{
	ARG_NONE,
	ARG_BOOL,
	ARG_INT,
	ARG_UINT,
	ARG_DOUBLE,
	ARG_DWORD,
// requires 'char*' ptr as target
	ARG_STRING,
	ARG_ZERO_INT,
};

typedef struct
{
// for name comparisons
	char arg_name[ARG_MAX_LENGHT];
// type of 'arg_dest' pointer
	int arg_type;
// pointer to destination
// not good, no warning if the variable type is not according
	void* arg_dest;
} arg_distributor_t;

arg_distributor_t args_distributing[] = 
{
//*only for warning echo
	{"notype", ARG_NONE, 0},
// no system tray icon
	{"nsti", ARG_ZERO_INT, &show_tray_icon},

// +++ registering settings +++
// multimedia timer period :) (if need less than 10ms time chunks - have to lower this; 0.001 - is 1ms, is minimum)
// 0 - for that stay normal OS value
	{"multimedia_timer_period", ARG_DOUBLE, &multimedia_timer_period},
// time for the initializing
	{"time_period__initializing", ARG_DOUBLE, &time_period__initializing},
// _short
	{"delay", ARG_DOUBLE, &time_period__initializing},
// registering time
	{"time_period__registering", ARG_DOUBLE, &time_period__registering},
// _short
	{"working", ARG_DOUBLE, &time_period__registering},
// shut down time
	{"time_period__finishing", ARG_DOUBLE, &time_period__finishing},

// which period to sleep (sec)
	{"sleep_time__wait", ARG_DOUBLE, &sleep_time__wait},
	{"sleep_time__register", ARG_DOUBLE, &sleep_time__register},
// _short
	{"sleep", ARG_DOUBLE, &sleep_time__register},

// use timers instead of the keyboard hotkeys
	{"use_timers", ARG_BOOL, &use_timers},
// write command changes current registering mode to pause
	{"write_pauses", ARG_BOOL, &write_pauses},
// clear command changes current registering mode to pause
	{"clear_pauses", ARG_BOOL, &clear_pauses},

	{"breakpoint_lists_load", ARG_UINT, &regdrag_breakpoints_allocate_num},
// +++ multithread +++
//// register not from one thread, but from all
//	{"register_from_all_threads", ARG_BOOL, &regdrag_register_from_all_threads},
// reuse thread array cells, for the case, if threads are often created/exited
	{"reusing_thread_pointers", ARG_BOOL, &regdrag_reusing_thread_pointers},
// input exact thread index for registering from
	{"register_from_thread_ix", ARG_UINT, &regdrag_register_from_thread_ix},
// register from few threads, or from all ("-1")
	{"register_from_threads_ixes", ARG_STRING, &regdrag_register_from_threads_str},
// suspend and resume a thread at each break moment (use what do you like, since the result seems differs only slightly)
	{"threads_suspend", ARG_BOOL, &regdrag_threads_suspend},
// --- multithread ---

// arguments, which must be used for exe
	{"args", ARG_STRING, &regdrag_process_args_arg},

	{"dir_watcher_working", ARG_STRING, &regdrag_dir_watcher_working_arg},

	{"dir_proc_exe", ARG_STRING, &regdrag_dir_process_exe_arg},

	{"dir_proc_working", ARG_STRING, &regdrag_dir_process_working_arg},
// --- registering settings ---
// +++ displaying settings +++
// how much dragging points to display
	{"points", ARG_UINT, &displaying_dragging_addresses_num},
// == close positioned address points (blocks) ==
// how much dragging blocks to display
	{"blocks", ARG_UINT, &displaying_dragging_blocks_num},
// how close placed points to unite 
// *distance is to one side
	{"distance_unite", ARG_UINT, &distance_unite},
// == functions ==
// how much dragging functions to display
	{"functions", ARG_UINT, &displaying_dragging_functions_num},
// == subsettings ==
// like 'distance_unite', but is only for displaying
//	{"additional_distance_show", ARG_UINT, &additional_distance_show},
// == dumping ==
// module names list for :  dump boolean mask of hits     -     0 - no hits occure, 1 - hit(s) occure
// "*" to dump all; ';' is separator; avoid of spaces or extensions
	{"dump_boolean_masks", ARG_STRING, &displaying_dump_boolean_masks},
// --- displaying settings ---
// +++ process priority +++
// this configuration is optimal, and some other will be necessary only very rare
// don't use high level priority, because it blocks work of kernel drivers, so they work pulsive
// watcher priority very needs to be higher than watching target
// -2 - Low, -1 - Below Normal, 0 - Normal, 1 - Above Normal, 2 - High, 3 - Realtime
	{"process_priority__watcher", ARG_INT, &process_priority__watcher},
	{"process_priority__exe", ARG_INT, &process_priority__exe},
// --- process priority ---
//	{"compile_info_path",ARG_STRING,&compile_info_path},
// *must not be arg - at all
//	{"code_offset",ARG_UINT,&code_offset},

// "Win" key is not available on all keyboards; change then
	{"hotkeys_modifier_change", ARG_BOOL, &hotkeys_modifier_change},

	{"break_on_exception_inside_process", ARG_INT, &break_on_exception_inside_process},

	{"break_on_exception_outside_process", ARG_INT, &break_on_exception_outside_process},
// *switch it off, if there is enough initialization time for the process
	{"events_resets_timer", ARG_INT, &events_resets_timer},
// how much exceptions to accept not breaking the process
	{"exception_events_max", ARG_INT, &exception_events_max},
//	{"read_map_files", ARG_BOOL, &read_map_files},
	{"read_map_main_exe", ARG_BOOL, &read_map_main_exe},
	{"read_map_loaded_dlls", ARG_BOOL, &read_map_loaded_dlls},
};

// *such defines can't be moved outside
#define ARG_NAMES_NUM (sizeof(args_distributing) / sizeof(*args_distributing))


int Args_ParseArg (int arg_num)
{
char arg[ARG_MAX_LENGHT];
int num = 1;
int i;
double value_double;
int value_int;

	strcpy (arg,Args[arg_num]);

	if (arg[0] == '/')
		arg[0] = '-';

	if (arg[0] != '-')
		Error ("not a switch :  \"%s\" (argument #%d in command line)\n",Args[arg_num],arg_num);

	for (i = 0;i < ARG_NAMES_NUM;i++)
		if (! _stricmp (&arg[1],args_distributing[i].arg_name))
			break;

	if (i == ARG_NAMES_NUM)
		Error ("unknown argument #%d in command line :  \"%s\"\n",arg_num,Args[arg_num]);

	switch (args_distributing[i].arg_type)
	{
	case ARG_NONE:
		LogFile_Warningf ("argument #%d / \"%s\" in command line - curently not supported.\n",arg_num,Args[arg_num],args_distributing[i].arg_type);
		goto no_value;
		break;
	case ARG_BOOL:
		{
			if (sscanf (Args[arg_num + 1], "%i", &value_int) != 1)
				goto value_error;
			*(bool*)args_distributing[i].arg_dest = !! value_int;
		}
		break;
	case ARG_INT:
		{
			if (sscanf (Args[arg_num + 1], "%i", &value_int) != 1)
				goto value_error;
			*(int*)args_distributing[i].arg_dest = value_int;
		}
		break;
	case ARG_UINT:
		{
			if (sscanf (Args[arg_num + 1], "%i", &value_int) != 1)
				goto value_error;
			*(uint*)args_distributing[i].arg_dest = (uint)value_int;
		}
		break;
	case ARG_DOUBLE:
		{
			if (sscanf (Args[arg_num + 1], "%lf", &value_double) != 1)
				goto value_error;
			*(double*)args_distributing[i].arg_dest = value_double;
		}
		break;
	case ARG_DWORD:
		{
			if (sscanf (Args[arg_num + 1], "%i", &value_int) != 1)
				goto value_error;
			*(DWORD*)args_distributing[i].arg_dest = (DWORD)value_int;
		}
		break;
	case ARG_STRING:
		{
// **, but does anyone care :)
			*(char**)args_distributing[i].arg_dest = _strdup (Args[arg_num + 1]);
		}
		break;
	case ARG_ZERO_INT:
		{
			*(int*)args_distributing[i].arg_dest = 0;
			goto no_value;
		}
		break;
	default:
		Error ("Program error - wrong argument #%d in command line \"%s\" type :  %d.\n",arg_num,Args[arg_num],args_distributing[i].arg_type);
		break;
	}

	num++;

no_value:

return num;

value_error:
	Error ("can't read value \"%s\" for argument #%d / \"%s\" in command line.\n",Args[arg_num + 1],arg_num,Args[arg_num]);

return 0;
}

int Args_ParseCmdLine (LPSTR lpCmdLine)
{
int argc = 0;
int i = 0;
int arg_i;
int in_quote;
bool arg1_quoted = 0;

// hmmm, for what this :)
	Strncpy (Args[argc++], main_arg, sizeof(Args[0]));

	while (lpCmdLine[i])
	{
		if (argc >= ARGS_MAX)
			Error("To many arguments (> %d) in command line.",ARGS_MAX);

		if (lpCmdLine[i] == ' ')
// skip space
		{
			i++;
			continue;
		}

// argument begins

		in_quote = lpCmdLine[i] == '\"';

		if (in_quote)
		{
			if (argc == 1)
				arg1_quoted = 1;
// skip beginning quote
			i++;
		}

		arg_i = 0;

		while ((in_quote || lpCmdLine[i] != ' ') && lpCmdLine[i] != '\"')
		{
			if (lpCmdLine[i] == '\0')
			{
				if (in_quote)
					Error ("Argument #%d do not have closing quote in command line.", argc);
				else
					break;
			}

			if (arg_i >= (ARG_MAX_LENGHT - 1))
				Error ("Too long argument (> %d symbols) #%d in command line.",ARG_MAX_LENGHT - 1,argc);

			Args[argc][arg_i++] = lpCmdLine[i++];
		}

		if (in_quote)
// skip finishing quote
			i++;

		Args[argc][arg_i] = 0;

		argc++;
	}

	args_num = argc;

	if (args_num <= 1)
		Error ("No arguments.");

// no, no way to validate arg1
#if 0
	i = 0;
	char symbol;
	while (symbol = Args[1][i])
	{
		if (symbol == '')
			Error ("Filename argument \"%s\" have invalid symbols.", Args[1]);
		i++;
	}
#endif
// except this one :)
	if (! arg1_quoted)
		Error ("Filename argument \"%s\" goes not in quotes, possible mess.", Args[1]);

	if (sizeof(regdrag_process_executable_file_name) >= strlen (Args[1]))
		strcpy (regdrag_process_executable_file_name, Args[1]);
	else
		Error ("Too long filename argument (> %d symbols) #1 in command line.", sizeof(regdrag_process_executable_file_name));

	for (i = 2;i < args_num;)
	{
		i += Args_ParseArg (i);
	}

	return argc;
}

