// class is by Alexander Ilyin
// *original belongs at the Drag Watcher project

// *for MSVC 6.0
#if _MSC_VER < 1300
#define WINVER 0x500
#endif
#include "IndicatorWnd.h"
#include "error.h"
#include "string.h"
#include "reg_drag.h"
//#include <sstream>
#include "string2.h"


// *for MSVC 6.0 it must be done by this way,
// plus it makes compatibility between different Windows
//static bool user32_functions_initialized = 0;
static bool functions_pointers_initialized = 0;
// this is something global :/
//typedef BOOL WINAPI BlockInput_t (BOOL);
//BlockInput_t* pfnBlockInput = 0;
static void (WINAPI* pfnSetLayeredWindowAttributes) (HWND hwnd, COLORREF crKey, BYTE bAlpha, DWORD dwFlags) = 0;
//typedef void (WINAPI* SetLayerdWindowAttributesFunc_t) (HWND, COLORREF, BYTE, DWORD);
//SetLayerdWindowAttributesFunc_t pfnSetLayeredWindowAttributes = (SetLayerdWindowAttributesFunc_t)-1;

// some info
// 'SetWindowDisplayAffinity ()' - to set where the window can be displayed


void CIndicatorWnd::InitFunctions ()
{
HMODULE h_user32;

// *"user32.dll" must be loaded before, for this way
	h_user32 = GetModuleHandle (TEXT("USER32"));

	if (h_user32)
	{
		*(FARPROC*)&pfnSetLayeredWindowAttributes = GetProcAddress (h_user32, "SetLayeredWindowAttributes");
//		*(FARPROC*)&pfnBlockInput = GetProcAddress (h_user32, "BlockInput");
	}

	functions_pointers_initialized = 1;
}

//CIndicatorWnd::CIndicatorWnd (HINSTANCE h_instance)
CIndicatorWnd::CIndicatorWnd ()
{
// == subclasses ==
	mc_hot_keys = 0;

	mh_inst = 0;
// == window ==
	mh_wnd = 0;
//	mh_menu = 0;
// == style ==
//	topmost = 0;
// == text buffer ==
//	txt_buf_size = 0;
//	text_buf = 0;
//	text_buf_new = 0;
	m_text_out.erase();

	if (! functions_pointers_initialized)
		InitFunctions ();
}

static LPCTSTR class_name = "IndicatorWndClass";

//bool CIndicatorWnd::Init (HINSTANCE h_instance, int window_mode, int text_buffer_size)
//bool CIndicatorWnd::Init (HINSTANCE h_instance, int window_mode)
bool CIndicatorWnd::Init (HINSTANCE h_instance)
{
	bool result;

	mh_inst = h_instance;

// invalid modes
	m_mode_interaction = IW_STYLE_NONE;
	m_mode_output = IW_MODE_NONE;

//	txt_buf_size = text_buffer_size;
//	text_buf = (char*)malloc (txt_buf_size);
//	text_buf_new = (char*)malloc (txt_buf_size);
//	*text_buf = 0;
//	*text_buf_new = 0;
	m_text_out.erase();

	result = WindowCreate ();

	mc_hot_keys = new CHotKeys (mh_wnd);

//	mc_wnd_menu = new CWndMenu;

	return result;
}

void CIndicatorWnd::InitTemplate ()
{
	CWndMenu* submenu = mc_wnd_menu.Submenu_Create (0, "Window Properties");
	wm_item_t item;
	item.type = WMENU_ITEM_TYPE__CHECKABLE;
	item.c_wnd_menu = 0;
	item.variable.class_ptr = this;

	item.text = "Transparent";
	item.variable.function = VariableUpdate;
	item.variable.ptr = &m_settings.transparent;
	submenu->Item_Insert (0, &item);
	item.text = "Always On Top";
	item.variable.function = VariableUpdate;
	item.variable.ptr = &m_settings.on_top;
	submenu->Item_Insert (1, &item);
	item.text = "No Taskbar Icon";
	item.variable.function = VariableUpdate;
	item.variable.ptr = &m_settings.no_taskbar_icon;
	submenu->Item_Insert (2, &item);
	item.text = "Accept Drag-And-Drop";
	item.variable.function = VariableUpdate;
	item.variable.ptr = &m_settings.drag_and_drop;
	submenu->Item_Insert (3, &item);
}

void CIndicatorWnd::Messaging ()
{
MSG msg;

//	if (PeekMessage (&msg, 0, 0, 0, PM_NOREMOVE))
	if (PeekMessage (&msg, 0, 0, 0, PM_REMOVE))
	{
// remove this message
//		GetMessage (&msg, 0, 0, 0);
		TranslateMessage (&msg);
		DispatchMessage (&msg);
	}
}

void CIndicatorWnd::Text_Put (char* text)
{
//	Strncpy (text_buf, text, txt_buf_size);
	m_text_out = text;
	RedrawWindow (mh_wnd, 0, 0, RDW_INVALIDATE);
}

// how window will interact with events, for example - with mouse cursor
void CIndicatorWnd::SetMode_Interaction (int mode)
{
//	BOOL B_result;

	m_mode_interaction = mode;
	switch (mode)
	{
	case IW_STYLE_PASSIVE:
//		if (pfnBlockInput)
//			B_result = pfnBlockInput (1);
// *same result is with WS_DISABLED
//		B_result = EnableWindow (mh_wnd, 0);
		break;
	case IW_STYLE_ACTIVE:
//		if (pfnBlockInput)
//			B_result = pfnBlockInput (0);
//		B_result = EnableWindow (mh_wnd, 1);
		break;
	default:
		Error ("CIndicatorWnd::SetModes () :  unknown or unsupported mode %d", mode);
	}
}

// what kind of data to output
void CIndicatorWnd::SetMode_Output (int mode)
{
// only these are supported atm
	Assert(mode == IW_MODE_STRING || mode == IW_MODE_BITMAP);
	m_mode_output = mode;
}

void CIndicatorWnd::WindowUpdate ()
{
	Assert(mh_wnd);
	SetAttributes ();
	ShowWindow (mh_wnd, SW_SHOWNORMAL);
	UpdateWindow (mh_wnd);
}


// ==== private ====

#ifndef WS_EX_NOACTIVATE
#define WS_EX_NOACTIVATE        0x08000000L
#endif

bool CIndicatorWnd::WindowCreate ()
{
	WNDCLASSEX wc;

	wc.cbSize = sizeof(wc);
//	wc.style = CS_DBLCLKS | CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
//	wc.style = CS_DBLCLKS;
	wc.style = 0;
	wc.lpfnWndProc = WndProcReq;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = mh_inst;
//	wc.hIcon = LoadIcon(hInst, MAKEINTRESOURCE(D2_ICON));
	wc.hIcon = 0;
	wc.hCursor = LoadCursor (0, IDC_ARROW);
//	wc.hbrBackground = (HBRUSH)GetStockObject (BLACK_BRUSH);
	wc.hbrBackground = (HBRUSH)GetStockObject (DKGRAY_BRUSH);
	wc.lpszMenuName = 0;
//	wc.lpszMenuName = MAKEINTRESOURCE(IDR_MENU1);
	wc.lpszClassName = class_name;
	wc.hIconSm = 0;

	if (! RegisterClassEx (&wc))
		return 0;

// *does anyone needs 'WS_OVERLAPPEDWINDOW', - I don't
// WS_EX_LAYERED - nonrectangular windows
// WS_EX_COMPOSITED - draw own layers ordered
// WS_EX_NOACTIVATE - some z order positioning passivity
	mh_wnd = CreateWindowEx
	(
//		WS_EX_APPWINDOW | (topmost ? WS_EX_TOPMOST : 0),
//		WS_EX_APPWINDOW,
//		WS_EX_TOOLWINDOW | (0 ? WS_EX_ACCEPTFILES : 0),
		0,
//		WS_EX_STATICEDGE,
//		WS_EX_NOACTIVATE,
		class_name,
		WINAPP_NAME,
// it is harder to catch sizing border when it is 'WS_THICKFRAME' (3 lines) without 'WS_BORDER' (1 line more)
// *WS_POPUP, which removes caption, must be optional I think
		WS_POPUP | WS_BORDER | WS_THICKFRAME,
// WS_DISABLED disables window moving and mouse right click, but messages still aren't coming through
//		WS_POPUP | WS_BORDER | WS_THICKFRAME | WS_DISABLED,
//		WS_DISABLED,
//		0,
//		WS_BORDER | WS_THICKFRAME,
// *it will be resized
		0,
		0,
		0,
		0,
		0, 
		0, 
		mh_inst, 
		0
	);

	Assert(mh_wnd);

// put pointer to class
// *use old function, for older than MSVC 7.0 compilers
#if _MSC_VER < 1300
	SetWindowLong (mh_wnd, GWL_USERDATA, (LONG)this);
#else
	SetWindowLongPtr (mh_wnd, GWLP_USERDATA, (LONG_PTR)this);
#endif

//	h_menu = GetMenu (mh_wnd);

	return 1;
}

//void CIW_VariableUpdate (void* variable_class_ptr)
void CIndicatorWnd::VariableUpdate (void* variable_class_ptr)
{
	CIndicatorWnd* c_indicator_wnd = (CIndicatorWnd*)variable_class_ptr;

	c_indicator_wnd->SetAttributes ();
}

void CIndicatorWnd::Draw_Text (HDC hdc)
{
	stringstream strm;
	HFONT hfont;
	string str;
	uint y;
	uint font_height;

	font_height = 13;
	hfont = CreateFont (font_height, 0, 0, 0, 400,
		0, 0, 0, ANSI_CHARSET, OUT_CHARACTER_PRECIS, CLIP_CHARACTER_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH,
		"Verdana");
	if (hfont)
		SelectObject (hdc, hfont);

	SetBkColor (hdc, RGB(64, 64, 64));
	SetTextColor (hdc, RGB(192, 192, 192));

//	strm << text_buf;
//	TextOut (hdc, 10, 10, text_buf, strlen(text_buf));
	strm << m_text_out;
	y = 10;
// test
	str = m_text_out;
	while (1)
	{
//		strm >> str;
		getline(strm, str);
		if (strm.fail())
		{
			Assert(strm.eof());
			break;
		}
		TextOut (hdc, 10, y, str.c_str(), str.length());
// font height already have space included
		y += font_height;
	}
}

void CIndicatorWnd::Draw_Bitmap (HDC hdc)
{
}

// *support begins only from Windows 2000
#define LWA_ALPHA 2
#define WS_EX_LAYERED 0x00080000

/*
void CIndicatorWnd::SetAttributes ()
{
	BOOL B_result;

	B_result = SetWindowPos (mh_wnd, settings.on_top ? HWND_TOPMOST : HWND_NOTOPMOST, settings.pos_x, settings.pos_y, settings.size_x, settings.size_y, SWP_FRAMECHANGED);

	DWORD style = 0;
//	style = GetWindowLong (mh_wnd, GWL_EXSTYLE);
	if (settings.no_taskbar_icon)
		style |= WS_EX_TOOLWINDOW;
	if (settings.drag_and_drop)
		style |= WS_EX_ACCEPTFILES;
	if (settings.transparent)
		style |= WS_EX_LAYERED;

	B_result = !! SetWindowLong (mh_wnd, GWL_EXSTYLE, style);

	if (pfnSetLayeredWindowAttributes)
	{
// Set 'WS_EX_LAYERED' on this window
//		SetWindowLong (mh_wnd, GWL_EXSTYLE, GetWindowLong (mh_wnd, GWL_EXSTYLE) | WS_EX_LAYERED);
// set window alpha
		pfnSetLayeredWindowAttributes (mh_wnd, 0, settings.transparency >> 24, LWA_ALPHA);
	}
}
*/

void CIndicatorWnd::SetAttributes ()
{
	BOOL B_result;

	DWORD style = 0;
//	style = GetWindowLong (mh_wnd, GWL_EXSTYLE);
	if (m_settings.no_taskbar_icon)
		style |= WS_EX_TOOLWINDOW;
	if (m_settings.drag_and_drop)
		style |= WS_EX_ACCEPTFILES;
	if (m_settings.transparent)
		style |= WS_EX_LAYERED;

	B_result = !! SetWindowLong (mh_wnd, GWL_EXSTYLE, style);

	if (pfnSetLayeredWindowAttributes)
	{
// Set 'WS_EX_LAYERED' on this window
//		SetWindowLong (mh_wnd, GWL_EXSTYLE, GetWindowLong (mh_wnd, GWL_EXSTYLE) | WS_EX_LAYERED);
// set window alpha
		pfnSetLayeredWindowAttributes (mh_wnd, 0, m_settings.transparency >> 24, LWA_ALPHA);
	}

// *by some reason, 'SetWindowPos ()' must go after 'SetWindowLong ()', otherwise - not always 'HWND_TOPMOST'
	B_result = SetWindowPos (mh_wnd, m_settings.on_top ? HWND_TOPMOST : HWND_NOTOPMOST, m_settings.pos_x, m_settings.pos_y, m_settings.size_x, m_settings.size_y, SWP_FRAMECHANGED);
}

LRESULT CALLBACK CIndicatorWnd::WndProc (HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	RECT rect;
	LRESULT lresult;

	switch (msg)
	{
	case WM_CREATE:
		break;

	case WM_PAINT:
		PAINTSTRUCT ps;
		HDC hdc;
		hdc = BeginPaint (hwnd, &ps);
		switch (m_mode_output)
		{
		case IW_MODE_STRING:
//		case IW_MODE_TEXT:
			Draw_Text (hdc);
			break;
		case IW_MODE_BITMAP:
			Draw_Bitmap (hdc);
			break;
		default:
			Error ("CIndicatorWnd::WndProc (WM_PAINT) :  unknown or unsupported mode %d", m_mode_output);
		}
		EndPaint (hwnd, &ps);
		break;

#if 1
	case WM_NCHITTEST:
//		if (m_mode_interaction == IW_STYLE_PASSIVE)
//			goto pass_to_window_below;
		lresult = DefWindowProc (hwnd, msg, wparam, lparam);;
		if (lresult == HTCLIENT && GetAsyncKeyState (MK_LBUTTON) < 0)
// *I hope this cheat have no lacks
			lresult = HTCAPTION;
		return lresult;
#endif

	case WM_CONTEXTMENU:
// *not very good; not all windows wants this
//		if (m_mode_interaction == IW_STYLE_PASSIVE)
//			goto pass_to_window_below;
		{
			BOOL B_result;
//			int pos_x = GET_X_LPARAM(lparam);
//			int pos_y = GET_Y_LPARAM(lparam);
			int pos_x = LOWORD(lparam);
			int pos_y = HIWORD(lparam);

			if (at_least_Win2K)
				B_result = TrackPopupMenuEx
				(
					mc_wnd_menu.h_menu,
					TPM_LEFTALIGN | TPM_LEFTBUTTON,
					pos_x,
					pos_y,
					hwnd,
					0
				);
		}
		break;

	case WM_INITMENUPOPUP:
		mc_wnd_menu.Init ();
		break;

	case WM_UNINITMENUPOPUP:
		mc_wnd_menu.Free ();
		break;

	case WM_COMMAND:
		lresult = 0;
		break;

	case WM_MENUCOMMAND:
		CWndMenu_OnMenuCommand ((HMENU)lparam, wparam);
		break;

	case WM_HOTKEY:
		mc_hot_keys->OnHotKey (wparam);
		break;

	case WM_MOVE:
	case WM_SIZE:
// *'lparam' contains client data
// settings has been changed, - get offset on the screen
		GetWindowRect (hwnd, &rect);
		m_settings.pos_x = rect.left;
		m_settings.pos_y = rect.top;
		m_settings.size_x = rect.right - rect.left;
		m_settings.size_y = rect.bottom - rect.top;
		break;
	}

	return DefWindowProc (hwnd, msg, wparam, lparam);

// WS_DISABLED blocks messages; so must make own passer
// complete crap seems
#if 0
pass_to_window_below:
	HWND hwnd_next;
// *GetWindow () also can
	hwnd_next = GetNextWindow (hwnd, GW_HWNDNEXT);
	::SendMessage (hwnd_next, msg, wparam, lparam);
#endif
}

LRESULT CALLBACK CIndicatorWnd::WndProcReq (HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
// restore pointer to class
#if _MSC_VER < 1300
	CIndicatorWnd* c_indicator_wnd = (CIndicatorWnd*)GetWindowLong (hwnd, GWL_USERDATA);
#else
	CIndicatorWnd* c_indicator_wnd = (CIndicatorWnd*)GetWindowLongPtr (hwnd, GWLP_USERDATA);
#endif

	if (! c_indicator_wnd)
		return DefWindowProc (hwnd, msg, wParam, lParam);

	return c_indicator_wnd->WndProc (hwnd, msg, wParam, lParam);
}
