// class is by Alexander Ilyin
// menu, belonging to some window
// *original belongs at the Drag Watcher project

// *for MSVC 6.0
#if _MSC_VER < 1300
#define WINVER 0x500
#endif
#include "WndMenu.h"
#include "error.h"

#ifndef ULONG_PTR
#define ULONG_PTR DWORD
#endif


#define WNDMENU_ERRORS 1

#define WNDMENU_ITEMS_MAX 50


CWndMenu::CWndMenu ()
{
//	items = NULL;
	items = (wm_item_t*)malloc (WNDMENU_ITEMS_MAX);
	items_num = 0;

//	h_menu = NULL;
// *this handle must exist before 'CWndMenu::Init ()',
// that on 'WM_CONTEXTMENU' it could be used, when the menu is actually not initialized yet
//	h_menu = CreateMenu ();
	h_menu = CreatePopupMenu ();

	initialized = 0;

	MENUINFO menuinfo;
	menuinfo.cbSize = sizeof(MENUINFO);
	menuinfo.fMask = MIM_MENUDATA | MIM_STYLE;
	menuinfo.dwStyle = MNS_NOTIFYBYPOS;
	menuinfo.dwMenuData = (ULONG_PTR)this;

	BOOL B_result;

	B_result = SetMenuInfo (h_menu, &menuinfo);
}

CWndMenu::~CWndMenu ()
{
	free (items);
	DestroyMenu (h_menu);

}

bool CWndMenu::Item_Put (int position, wm_item_t* item_new)
{
	wm_item_t* item = &items[position];

	if (position < 0 || position >= items_num)
#if WNDMENU_ERRORS
		Error ("'CWndMenu::Item_Put ()' :  position (%d) is out of range (0-%d)", position, WNDMENU_ITEMS_MAX);
#else
		return 0;
#endif

	Assert(item_new->type != WMENU_ITEM_TYPE__BUTTON || item_new->variable.function || item_new->variable.ptr);
// check specific arguments
	Assert((item_new->type == WMENU_ITEM_TYPE__SEPARATOR) ^ (!! item_new->text));
	Assert((item_new->type == WMENU_ITEM_TYPE__SUBMENU) ^ (! item_new->c_wnd_menu));

	*item = *item_new;
	if (item_new->text)
		item->text = _strdup (item_new->text);
	else
		item->text = NULL;

	return 1;
}

bool CWndMenu::Item_Insert (int position, wm_item_t* item_new)
{
	int i;

	if (items_num >= WNDMENU_ITEMS_MAX)
#if WNDMENU_ERRORS
		Error ("'CWndMenu::Item_Insert ()' :  'WNDMENU_ITEMS_MAX' (%d) is reached", WNDMENU_ITEMS_MAX);
#else
		return 0;
#endif

	if (position < 0 || position > items_num)
#if WNDMENU_ERRORS
		Error ("'CWndMenu::Item_Insert ()' :  position (%d) is out of range (0-%d)", position, WNDMENU_ITEMS_MAX);
#else
		return 0;
#endif

	for (i = items_num - 1; i >= position; i--)
		items[i + 1] = items[i];

	items_num++;

	Item_Put (position, item_new);

	return 1;
}

CWndMenu* CWndMenu::Submenu_Create (int position, char* text)
{
	bool result;

	CWndMenu* c_wnd_menu = new CWndMenu;
	wm_item_t item;
	item.type = WMENU_ITEM_TYPE__SUBMENU;
	item.c_wnd_menu = c_wnd_menu;
	item.text = text;
	item.variable.function = NULL;
	item.variable.class_ptr = NULL;
	item.variable.ptr = NULL;

	result = CWndMenu::Item_Insert (position, &item);

	return c_wnd_menu;
}

bool CWndMenu::Item_Remove (int position)
{
	wm_item_t* item = &items[position];
	int i;

	if (position < 0 || position >= items_num)
#if WNDMENU_ERRORS
		Error ("'CWndMenu::Item_Insert ()' :  item at the position (%d) doesn't exist (0-%d are existing)", position, items_num);
#else
		return 0;
#endif

	if (item->text)
		free (item->text);

	if (item->c_wnd_menu)
		delete item->c_wnd_menu;

	items_num--;

	for (i = position; i < items_num; i++)
		items[i] = items[i + 1];

	return 1;
}

void CWndMenu::Item_RemoveAll ()
{
	int i;

	for (i = items_num - 1; i >= 0; i--)
		Item_Remove (i);
}


bool CWndMenu::Init ()
{
	bool result = 1;
	int i;
	UINT uFlags;
	UINT uIDNewItem;
	BOOL B_result;

	if (initialized)
		return 0;
	initialized = 1;

	for (i = 0; i < items_num; i++)
	{
	wm_item_t* item = &items[i];

		uFlags = 0;
		uIDNewItem = i;

		if (item->text)
			uFlags |= MF_STRING;

		switch (item->type)
		{
		case WMENU_ITEM_TYPE__GRAYED:
			uFlags |= MF_GRAYED;
			break;
		case WMENU_ITEM_TYPE__SUBMENU:
			uFlags |= MF_ENABLED | MF_POPUP;
			item->c_wnd_menu->Init ();
			uIDNewItem = (UINT)item->c_wnd_menu->h_menu;
			break;
		case WMENU_ITEM_TYPE__SEPARATOR:
			uFlags |= MF_SEPARATOR;
			break;
		case WMENU_ITEM_TYPE__BUTTON:
			uFlags |= MF_ENABLED;
			break;
		case WMENU_ITEM_TYPE__CHECKABLE:
			uFlags |= MF_ENABLED | (*(bool*)item->variable.ptr ? MF_CHECKED : 0);
			break;
		default:
			break;
		}

		B_result = AppendMenu
		(
			h_menu,
			uFlags,
			uIDNewItem,
			item->text
		);

		if (! B_result)
			result = 0;
	}

	return result;
}

void CWndMenu::Free ()
{
	int i;
	BOOL result;

	if (! initialized)
		return;
	initialized = 0;

// delete all items in this menu
// *no no no, the menu handle will be needed
//	result = DestroyMenu (h_menu);
//	h_menu = CreatePopupMenu ();

	for (i = items_num - 1; i >= 0; i--)
	{
	wm_item_t* item = &items[i];

		result = RemoveMenu (h_menu, i, MF_BYPOSITION);
		Assert(result);

		switch (item->type)
		{
		case WMENU_ITEM_TYPE__GRAYED:
			break;
		case WMENU_ITEM_TYPE__SUBMENU:
			if (! item->c_wnd_menu)
				Error ("'CWndMenu::Free ()' :  submenu is NULL", WNDMENU_ITEMS_MAX);
			item->c_wnd_menu->Free ();
			break;
		case WMENU_ITEM_TYPE__SEPARATOR:
		case WMENU_ITEM_TYPE__BUTTON:
		case WMENU_ITEM_TYPE__CHECKABLE:
			break;
		}
	}

// no reminder must be
	Assert(! RemoveMenu (h_menu, 0, MF_BYPOSITION));

//	h_menu = NULL;
}


void CWndMenu::OnMenuCommand (int position)
{
	wm_item_t* item = &items[position];

	switch (item->type)
	{
	case WMENU_ITEM_TYPE__GRAYED:
	case WMENU_ITEM_TYPE__SUBMENU:
	case WMENU_ITEM_TYPE__SEPARATOR:
		break;
	case WMENU_ITEM_TYPE__BUTTON:
		if (item->variable.ptr)
			*(bool*)item->variable.ptr = 1;
		if (item->variable.function)
			item->variable.function (item->variable.class_ptr);
		break;
	case WMENU_ITEM_TYPE__CHECKABLE:
		if (item->variable.ptr)
			*(bool*)item->variable.ptr = ! *(bool*)item->variable.ptr;
		if (item->variable.function)
			item->variable.function (item->variable.class_ptr);
		break;
	default:
		Error ("'CWndMenu::OnMenuCommand ()' :  wrong type");
	}
}

void CWndMenu_OnMenuCommand (HMENU hmenu, int position)
{
BOOL B_result;
MENUINFO menuinfo;
CWndMenu* c_menu;

	menuinfo.cbSize = sizeof(MENUINFO);
	menuinfo.fMask = MIM_MENUDATA;
	B_result = GetMenuInfo (hmenu, &menuinfo);
	if (! B_result)
		return;
	c_menu = (CWndMenu*)menuinfo.dwMenuData;
	c_menu->OnMenuCommand (position);
}
