// by Alexander Ilyin
// the main intention of this currently is only one :
// find out, what places are not in use
// *if you need watch dragging points - don't mess with this stuff, it normally is not in use, and hardly can be used for a debug purposes at all

#include "break_points.h"
//#include <stdlib.h>
#include "log_file.h"
#include "error.h"


//void BreakPoints::Init (uint points_num)
BreakPoints::BreakPoints (uint points_num)
{
	m_handle_process = 0;
//	m_break_points = 0;
	m_points_allocated_num = 0;
	m_points_num = 0;
	m_break_points = (break_point_t*)malloc (sizeof(break_point_t) * points_num);
	if (! m_break_points)
	{
		LogFile_Printf ("----- BreakPoints::Init () failed -----\n", points_num);
	}
	m_points_allocated_num = points_num;
	m_bp_last_address_enabled = 0;
	m_bp_last_address_disabled = 0;
}

//void BreakPoints::Quit ()
BreakPoints::~BreakPoints ()
{
	Assert(!! m_break_points);
	Assert(!! m_points_allocated_num);
	free (m_break_points);
	m_points_allocated_num = 0;
	m_handle_process = 0;
}

void BreakPoints::SetHandleProcess (HANDLE handle_process)
{
	Assert(! m_handle_process);
	m_handle_process = handle_process;
}

bool BreakPoints::ByteRead (ulong address, byte& b)
{
	BOOL B_result;
	byte buffer[64];
	DWORD num_bytes = -1;

	Assert(!! m_handle_process);

// get single byte at that position
	B_result = ReadProcessMemory (m_handle_process, (void*)address, buffer, 1, &num_bytes);
	if (! B_result)
	{
		LogFile_Printf ("----- BreakPoints::ByteRead (%08x) :  ReadProcessMemory () failed -----\n", address);
		return 0;
	}
	b = buffer[0];

	return 1;
}

bool BreakPoints::ByteWrite (ulong address, byte b)
{
	BOOL B_result;
	byte buffer[64];
	DWORD num_bytes = -1;

	Assert(!! m_handle_process);

// write single byte at that position
	buffer[0] = b;
	B_result = WriteProcessMemory (m_handle_process, (void*)address, buffer, 1, &num_bytes);
	if (! B_result)
	{
		LogFile_Printf ("----- BreakPoints::ByteWrite (%08x) :  WriteProcessMemory () failed -----\n", address);
		return 0;
	}

	return 1;
}

// put sorted
bool BreakPoints::Add (ulong address, uint* ix)
{
	bool result;
	uint i;
	uint pos;
	break_point_t* bp;
	byte b;

	Assert(!! m_handle_process);

	if (m_points_num >= m_points_allocated_num)
	{
		LogFile_Printf ("----- BreakPoints::Add (%08x) :  points number reached maximum (%u) -----\n", address, m_points_allocated_num);
		return 0;
	}

// search (sorting)
	for (i = 0; i < m_points_num; i++)
	{
		if (address == m_break_points[i].address)
		{
			LogFile_Printf ("----- BreakPoints::Add (%08x) :  points already exist -----\n", address);
			return 0;
		}
		if (address < m_break_points[i].address)
			break;
	}
//	LogFile_Printf ("----- BreakPoints::Add (%08x) :  WriteProcessMemory () failed -----\n", address);
//	return 0;
	pos = i;
	if (ix)
		*ix = pos;

// move those which are with higher addresses
	if (m_points_num)
		for (i = m_points_num - 1; i >= pos; i--)
		{
			m_break_points[i + 1] = m_break_points[i];
		}

// backup single byte at that position
	result = ByteRead (address, b);
	if (! result)
	{
		LogFile_Printf ("----- BreakPoints::Add (%08x) :  read failed -----\n", address);
		return 0;
	}
	if (b == 0xcc)
		LogFile_Printf ("- BreakPoints::Add (%08x) warning :  byte at address already is 0xcc (int 3) -\n", address);

//put_point:
	bp = &m_break_points[pos];
	bp->address = address;
// but no need now
	bp->byte_at_position = b;
	bp->is_on = 0;

	m_points_num++;

	return 1;
}

//bool BreakPoints::Delete (uint ix)
void BreakPoints::Delete (uint ix)
{
	break_point_t* bp;
	uint i;

	Assert(ix < m_points_num);

	bp = &m_break_points[ix];
	if (bp->is_on)
		Disable (ix);
	for (i = ix + 1; i < m_points_num; i++)
	{
		m_break_points[i - 1] = m_break_points[i];
	}
	m_points_num--;
}

bool BreakPoints::AddPointsByBooleanArray (ulong base_address, byte* array, uint size, uint& num_added)
{
	ulong address;
	bool result;
	uint i;
	uint ix;
//	uint num_added;

	address = base_address;
	i = 0;
	num_added = 0;
loop:
	if (! array[i])
		goto loop_continue;
	result = Add (address, &ix);
	if (! result)
		return 0;
	num_added++;
	result = Enable (ix);
loop_continue:
	address++;
	i++;
	if (i < size)
		goto loop;
	return 1;
}

//bool BreakPoints::IsOur (ulong address)
bool BreakPoints::Find (ulong address, uint& ix)
{
	uint i;

	for (i = 0; i < m_points_num; i++)
	{
		if (address == m_break_points[i].address)
		{
			ix = i;
			return 1;
		}
	}
	return 0;
}

//bool BreakPoints::Enable (ulong address)
bool BreakPoints::Enable (uint ix)
{
	bool result;
	byte b;
	break_point_t* bp;

	Assert(!! m_handle_process);
	Assert(ix < m_points_num);

	bp = &m_break_points[ix];
	if (bp->is_on)
	{
		LogFile_Printf ("----- BreakPoints::Enable (%08x) :  breakpoint at %08x already is enabled -----\n", ix, bp->address);
		return 0;
	}
	result = ByteRead (bp->address, b);
	if (! result)
	{
		LogFile_Printf ("----- BreakPoints::Enable (%08x) :  read failed -----\n", ix);
		return 0;
	}
	if (b == 0xcc)
	{
		LogFile_Printf ("----- BreakPoints::Enable (%08x) :  byte at address %08x is 0xcc, avoiding of changing it -----\n", ix, bp->address);
		return 0;
	}

	bp->byte_at_position = b;
	result = ByteWrite (bp->address, 0xcc);
	if (! result)
	{
		LogFile_Printf ("----- BreakPoints::Enable (%08x) :  write failed -----\n", ix);
		return 0;
	}
	bp->is_on = 1;
	m_bp_last_address_enabled = bp->address;

	return 1;
}

//bool BreakPoints::Disable (ulong address)
bool BreakPoints::Disable (uint ix)
{
	bool result;
	byte b;
	break_point_t* bp;

	Assert(!! m_handle_process);
	Assert(ix < m_points_num);

	bp = &m_break_points[ix];
	if (! bp->is_on)
	{
		LogFile_Printf ("----- BreakPoints::Disable (%08x) :  breakpoint at %08x already is disabled -----\n", ix, bp->address);
		return 0;
	}
	result = ByteRead (bp->address, b);
	if (! result)
	{
		LogFile_Printf ("----- BreakPoints::Disable (%08x) :  read failed -----\n", ix);
		return 0;
	}
	if (b != 0xcc)
	{
		LogFile_Printf ("----- BreakPoints::Disable (%08x) :  byte at address %08x is %02x but not 0xcc, avoiding of changing it -----\n", ix, bp->address, b);
		return 0;
	}

	b = bp->byte_at_position;
	result = ByteWrite (bp->address, b);
	if (! result)
	{
		LogFile_Printf ("----- BreakPoints::Disable (%08x) :  write failed -----\n", ix);
		return 0;
	}
	bp->is_on = 0;
	m_bp_last_address_disabled = bp->address;

	return 1;
}

// *not enabled are not intended to be printed atm in any case
void BreakPoints::DumpRemained_ToLog (bool only_enabled)
{
	char* list_name;
	uint i;
	uint j;
	break_point_t* bp;

	if (only_enabled)
		list_name = "remained breakpoints list";
	else
		list_name = "breakpoints list";
	LogFile_Printf ("+ %s +\n", list_name);
	for (i = 0, j = 0; i < m_points_num; i++)
	{
		bp = &m_break_points[i];
		if (! bp->is_on)
			if (only_enabled)
				continue;
// spacing
		if (j)
			LogFile_Printf ("   ");
		if (only_enabled)
			LogFile_Printf ("%08x", bp->address);
		else
//			LogFile_Printf ("%u %08x", bp->is_on, bp->address);
			LogFile_Printf ("%c %08x", bp->is_on ? '+' : '-', bp->address);
		j++;
		if (j >= 4)
		{
			j = 0;
//		if (! j)
			LogFile_Printf ("\n");
		}
	}
// finish line, if isn't finished
	if (j)
		LogFile_Printf ("\n");
	LogFile_Printf ("- %s -\n", list_name);
	LogFile_Printf ("\n");
}

