// by Alexander Ilyin
// display sorted debug results

#include "displaying.h"
#include <windows.h>
#include <assert.h>
#include <stdio.h>
#include "types.h"
#include "log_file.h"
#include "sorting.h"
#include "reg_drag.h"
#include "map_file.h"
//#include "string.h"
#include "error.h"
//#include <string>
//#include <sstream>
//using namespace std;


#define POINTS_MAX 10000
#define BLOCKS_MAX 1000
#define FUNCTIONS_MAX 1000


// +++ settings +++

// == main ==
// displaying granularity
int raw_volume = 0x10;
// dump all
bool full_dump = 0;
// dump hits as boolean mask; module names list
char* displaying_dump_boolean_masks = 0;

// == address points ==
// how much dragging points to display
int displaying_dragging_addresses_num = 25;

// == close positioned address points ==
// how much dragging blocks to display
int displaying_dragging_blocks_num = 25;

// == functions ==
// how much dragging functions to display
int displaying_dragging_functions_num = 25;

// == subsettings ==
// like 'distance_unite', but is only for displaying
int additional_distance_show = 0x08;

// --- settings ---


int Displaying_ShowFunction (CWatchImage* c_image, uint offset)
{
//int function_index;
map_file_function_info_t* map_file_function_info;
bool result;
char string_c_type[30];

	Assert(offset < 0x80000000);

	result = c_image->GetFunctionInfo (&map_file_function_info, offset);

	if (result)
	{
//	map_file_function_info_t* map_file_function_info = &c_map_file->map_file_functions_info[function_index];

		MapFile_GetFunctionTypeString (string_c_type, map_file_function_info->c_type);
//		LogFile_Printf ("Function \"%s\" %08x +%08x (%s)\n", map_file_function_info->function_name, map_file_function_info->beginning_address, offset - map_file_function_info->beginning_offset, map_file_function_info->obj_name);
		LogFile_Printf ("Function \"%s\" %s %08x +%08x (%s)\n", map_file_function_info->function_name, string_c_type, map_file_function_info->offset_in_image, offset - map_file_function_info->offset_in_image, map_file_function_info->obj_name);
	}
	else
	{
		LogFile_Printf ("No known function at offset %08x\n", offset);
		return 0;
	}

	return 1;
}

// **nnnooo, not in log
void Displaying_ExportRegionDump (CWatchImage* c_image)
{
int i;
uint j;
byte* displaying_block;
uint base_of_image = c_image->image_base;
uint image_size = c_image->image_size;
BOOL B_result;

	if (c_image->have_file_path)
		LogFile_Printf ("%S\n", c_image->file_path);
	LogFile_Printf ("base of image :  %08x\n", base_of_image);

	displaying_block = (byte*)malloc (image_size);

	B_result = ReadProcessMemory (regdrag_h_process, (void*)base_of_image, displaying_block, image_size, 0);

	for (j = 0; j < image_size;)
	{
// print the address at the beginning of the line
		LogFile_Printf ("%08x  ", j);

		for (i = 0; i < raw_volume; i++, j++)
		{
			if (i)
// separate with space
				LogFile_Printf (" ");
// print byte
			LogFile_Printf ("%02x", displaying_block[j]);
		}

// finish the line
		LogFile_Printf ("\n");
	}

	LogFile_Printf ("\n");

	free (displaying_block);
}

void Displaying_DumpBooleanMask (CWatchImage* image)
{
#if 0
	FILE* f;

	f = fopen (image->file_name, "rb");
	fclose (f);
#endif
	BOOL B_result;
	HANDLE h_file;
	char buf_file_name[1024];
	DWORD bytes_written = 0;
	byte* buffer_boolean_mask;
	uint size;
	uint i;

	sprintf (buf_file_name, "%s.dmpbm", image->file_name);
	h_file = CreateFile (buf_file_name, GENERIC_WRITE, FILE_SHARE_READ, 0, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
	if (h_file == INVALID_HANDLE_VALUE)
	{
		LogFile_Printf ("- CreateFile (%s) for write has been failed -\n", buf_file_name);
		return;
	}

// *fast decision way; very probable - is improper; but it is equal to method used in CWatchImage anyway
//	size = image->registration_array_size;
	size = image->image_size;
	buffer_boolean_mask = (byte*)malloc (size);
	if (! buffer_boolean_mask)
	{
		LogFile_Printf ("- malloc (%u) for module \"%s\" has been failed -\n", size, buf_file_name);
		goto close_file;
	}

	for (i = 0; i < size; i++)
	{
		buffer_boolean_mask[i] = !! image->registration_array[i];
	}

	B_result = WriteFile (h_file, buffer_boolean_mask, size, &bytes_written, 0);
	if (B_result == 0)
		LogFile_Printf ("- WriteFile (%s) has been failed -\n", buf_file_name);

	free (buffer_boolean_mask);

close_file:
	B_result = CloseHandle (h_file);
}

void Displaying_Process_DumpBooleanMasks ()
{
	int i;
	bool mark_all = 0;
	int dumped_num = 0;
	char* ptr;
	char* ptr_separator;
	int name_length;
	bool found;

	Assert(!! displaying_dump_boolean_masks);

	LogFile_Printf ("Dumping boolean masks.\n");

	if (*displaying_dump_boolean_masks == '*')
		mark_all = 1;
	ptr = displaying_dump_boolean_masks;
scan_loop:
	if (! *ptr)
		goto dump_loop;
	ptr_separator = strchr (ptr, ';');
	if (ptr_separator)
		name_length = ptr_separator - ptr;
	else
		name_length = strlen (ptr);
	found = 0;
	for (i = 0; i < images_num; i++)
	{
		CWatchImage* img = c_images[i];
		bool equal;

		if (mark_all)
		{
			img->displaying_dump_boolean_mask = 1;
			continue;
		}
		equal = ! _strnicmp (ptr, img->file_name, name_length);
		if (equal)
		{
			img->displaying_dump_boolean_mask = 1;
			found = 1;
		}
	}
	if (mark_all)
		goto dump_loop;
	if (! found)
		LogFile_Printf ("- module \"%.*s\" not found -\n", name_length, ptr);
	if (ptr_separator)
	{
		ptr = ptr_separator;
// skip separator
		ptr++;
		goto scan_loop;
	}

dump_loop:
	for (i = 0; i < images_num; i++)
	{
		CWatchImage* img = c_images[i];

		if (! img->displaying_dump_boolean_mask)
			continue;
		Displaying_DumpBooleanMask (img);
		dumped_num++;
	}

	LogFile_Printf ("%u boolean mask(s) dumped.\n", dumped_num);
	LogFile_Printf ("\n");
}

void Displaying_ExportResults ()
{
int i;
//uint j;
uint size;
CWatchImage* c_image;
uint* registration_array;

	LogFile_Printf ("Writing results.\n\n");

// + total +
	int timer_registering_milliseconds;
	int timer_registering_seconds;
	double sleep_time = 0;
	if (total_hits)
		sleep_time = (double)timer_registering / total_hits;
	timer_registering_milliseconds = timer_registering % 1000;
	timer_registering_seconds = timer_registering / 1000;
//	if (regdrag_register_from_all_threads)
//		LogFile_Printf ("Notice :  since 'register_from_all_threads' is set, the first two values are representing hits for all threads\n");
	if (regdrag_register_from_threads_num)
	{
		if (regdrag_register_from_threads_num == -1)
			LogFile_Printf ("Notice :  since \"register from all threads\" is set, the first two values are representing hits for all threads\n");
		else
			LogFile_Printf ("Notice :  since \"register from multiple threads\" is set, the first two values are representing hits for all threads\n");
// display all threads' indices
		string str;
		RegDrag_PrintThreadInfos (str);
		LogFile_Printf ("Threads (index, id) :  %s\n", str.c_str());
	}
	LogFile_Printf ("Breaking inside of process :  %I64d\n", inside_process);
	LogFile_Printf ("Breaking outside of process :  %I64d\n", outside_process);
	LogFile_Printf ("Total hits :  %I64d\n", total_hits);
	LogFile_Printf ("Registering time :  %d.%03d\n", timer_registering_seconds, timer_registering_milliseconds);
	LogFile_Printf ("Average sleep time :  %.3fms (%.3fs is in settings)\n", sleep_time, sleep_time__register);
	LogFile_Printf ("\n");
// - total -

// + per region +
	size = IMAGES_MAX * sizeof(*modules);
	memset (modules, 0xff, size);

	Sorting_ByModules ();

	LogFile_Printf ("Breaking inside of modules :\n");

#define MODULE_NAME (c_image->have_file_path ? c_image->file_name : "unknown")

	for (i = 0; i < IMAGES_MAX; i++)
	{
		module_info_t* module = &modules[i];

		if (module->image_index < 0)
		{
			LogFile_Printf ("- No more modules -\n");
			break;
		}

		c_image = c_images[module->image_index];

		LogFile_Printf ("module \"%s\", %016I64x times\n", MODULE_NAME, c_image->hits_counter);
	}
	LogFile_Printf ("\n");
// - per region -

	num_points = min(POINTS_MAX,displaying_dragging_addresses_num);
	num_blocks = min(BLOCKS_MAX,displaying_dragging_blocks_num);
	num_functions = min(FUNCTIONS_MAX,displaying_dragging_functions_num);

//// *for not used object, the 'image_index' member will be -1
// --------------------------------------------------------------------
// *for not used object, the 'number' count value will be 0
// --------------------------------------------------------------------

	size = num_points * sizeof(*points);
	points = (point_info_t*)malloc (size);
//	ZeroMemory (points,size);
//	memset (points, 0xff, size);
	memset (points, 0, size);

	size = num_blocks * sizeof(block_info_t);
	blocks = (block_info_t*)malloc (size);
//	memset (blocks, 0xff, size);
	memset (blocks, 0, size);

	size = num_functions * sizeof(function_info_t);
	functions = (function_info_t*)malloc (size);
//	memset (functions, 0xff, size);
	memset (functions, 0, size);


	if (num_points)
	{
		Sorting_ByPoints ();

		LogFile_Printf ("Mostly dragging points :\n");

		if (points[0].hits_number == 0)
		{
			LogFile_Printf ("Here were no points, that can be displayed.\n");
			goto no_points;
		}

		for (i = 0; i < num_points; i++)
		{
		point_info_t* point = &points[i];

//			if (point->image_index < 0)
			if (point->hits_number == 0)
			{
				LogFile_Printf ("- No more points -\n");
				break;
			}

			c_image = c_images[point->image_index];
			registration_array = c_image->registration_array;

			LogFile_Printf ("point %02x, %08x times, module \"%s\", address %08x\n", i, registration_array[point->address], MODULE_NAME, point->address);
		}

no_points:
		LogFile_Printf ("\n");
	}

	if (num_blocks)
	{
		Sorting_ByBlocks ();

		LogFile_Printf ("Mostly dragging blocks :\n");

//		if (blocks[0].image_index < 0)
		if (blocks[0].number == 0)
		{
			LogFile_Printf ("Here were no points in blocks, that can be displayed.\n");
			goto noblocks;
		}

		for (i = 0; i < num_blocks; i++)
		{
		block_info_t* block = &blocks[i];

//			if (block->image_index < 0)
			if (block->number == 0)
// end of info
			{
				LogFile_Printf ("- No more blocks -\n");
				break;
			}

			c_image = c_images[block->image_index];
			registration_array = c_image->registration_array;

			LogFile_Printf ("block %02x, module \"%s\", begins %08x, ends %08x  -  %016I64x times\n", i, MODULE_NAME, block->start, block->end, block->number);
			LogFile_Printf ("     mostly dragging point of block is %08x, %08x times\n", block->mostly, block->number_for_mostly);
		}

noblocks:
		LogFile_Printf ("\n");
	}


	if (num_functions)
	{
		Sorting_ByFunctions ();

		LogFile_Printf ("Mostly dragging functions :\n");

		if (functions[0].number == 0)
		{
			LogFile_Printf ("Here were no points in functions, that can be displayed.\n");
			goto nofunctions;
		}

		for (i = 0; i < num_functions; i++)
		{
		function_info_t* function = &functions[i];
		char string_c_type[30];

			if (function->number == 0)
// end of info
			{
				LogFile_Printf ("- No more functions -\n");
				break;
			}

//			c_image = c_images[function->image_index];
			c_image = function->c_image;
			registration_array = c_image->registration_array;
//			map_file_function_info_t* map_file_function_info = &c_map_file->map_file_functions_info[function->function_index];
			map_file_function_info_t* map_file_function_info = function->map_file_function_info;
			MapFile_GetFunctionTypeString (string_c_type, map_file_function_info->c_type);

			LogFile_Printf ("function %02x, module \"%s\", name \"%s\" %s, object \"%s\", offset %08x  -  %016I64x times\n", i, MODULE_NAME, map_file_function_info->function_name, string_c_type, map_file_function_info->obj_name, map_file_function_info->offset_in_image, function->number);
			LogFile_Printf ("     mostly dragging point of function is at + %08x, %08x times\n", function->offset_mostly, function->number_for_mostly);
		}
//		else
//			LogFile_Printf ("--- Can't find out mostly dragging functions - \"%s\" file isn't loaded. ---\n",map_file_path);

nofunctions:
		LogFile_Printf ("\n");

//		delete c_map_file;
//		c_map_file = 0;
	}


	if (full_dump)
	{
		LogFile_Printf ("Process execution memory (full dump) :\n\n");

		for (i = 0; i < images_num; i++)
			Displaying_ExportRegionDump (c_images[i]);

		LogFile_Printf ("\n");
	}

	if (displaying_dump_boolean_masks)
		Displaying_Process_DumpBooleanMasks ();

	LogFile_Printf ("\nEnd of results.\n\n");

	if (points)
		free (points);
	if (blocks)
		free (blocks);
	if (functions)
		free (functions);
}

