// by Alexander Ilyin
// NT filesystem access, using internal NT names
// *original belongs to scanner with ClamAV library usage

#include "file_nt.h"
//#define _NTDDK_
//#include <ks.h>
//#include <winnt.h>
#include <winioctl.h>


//typedef LONG NTSTATUS;

extern "C"
{
typedef
NTSYSAPI
NTSTATUS
NTAPI
ZwOpenFile_t(
    OUT PHANDLE FileHandle,
    IN ACCESS_MASK DesiredAccess,
    IN POBJECT_ATTRIBUTES ObjectAttributes,
    OUT PIO_STATUS_BLOCK IoStatusBlock,
    IN ULONG ShareAccess,
    IN ULONG OpenOptions
    );

typedef
NTSYSAPI
NTSTATUS
NTAPI
ZwReadFile_t(
    IN HANDLE FileHandle,
    IN HANDLE Event OPTIONAL,
    IN PIO_APC_ROUTINE ApcRoutine OPTIONAL,
    IN PVOID ApcContext OPTIONAL,
    OUT PIO_STATUS_BLOCK IoStatusBlock,
    OUT PVOID Buffer,
    IN ULONG Length,
    IN PLARGE_INTEGER ByteOffset OPTIONAL,
    IN PULONG Key OPTIONAL
    );

typedef
NTSYSAPI
NTSTATUS
NTAPI
ZwClose_t(
    IN HANDLE Handle
    );

typedef
NTSYSAPI
NTSTATUS
NTAPI
ZwQueryInformationFile_t(
    IN HANDLE FileHandle,
    OUT PIO_STATUS_BLOCK IoStatusBlock,
    OUT PVOID FileInformation,
    IN ULONG Length,
    IN FILE_INFORMATION_CLASS FileInformationClass
    );
}

ZwOpenFile_t* ZwOpenFile = 0;
ZwReadFile_t* ZwReadFile = 0;
ZwClose_t* ZwClose = 0;
ZwQueryInformationFile_t* ZwQueryInformationFile = 0;


#define FILE_NON_DIRECTORY_FILE                 0x00000040

#define InitializeObjectAttributes( p, n, a, r, s ) { \
    (p)->Length = sizeof( OBJECT_ATTRIBUTES );          \
    (p)->RootDirectory = r;                             \
    (p)->Attributes = a;                                \
    (p)->ObjectName = n;                                \
    (p)->SecurityDescriptor = s;                        \
    (p)->SecurityQualityOfService = NULL;               \
    }

#if 0
typedef enum _FILE_INFORMATION_CLASS {
// end_wdm
    FileDirectoryInformation         = 1,
    FileFullDirectoryInformation,   // 2
    FileBothDirectoryInformation,   // 3
    FileBasicInformation,           // 4  wdm
    FileStandardInformation,        // 5  wdm
    FileInternalInformation,        // 6
    FileEaInformation,              // 7
    FileAccessInformation,          // 8
    FileNameInformation,            // 9
    FileRenameInformation,          // 10
    FileLinkInformation,            // 11
    FileNamesInformation,           // 12
    FileDispositionInformation,     // 13
    FilePositionInformation,        // 14 wdm
    FileFullEaInformation,          // 15
    FileModeInformation,            // 16
    FileAlignmentInformation,       // 17
    FileAllInformation,             // 18
    FileAllocationInformation,      // 19
    FileEndOfFileInformation,       // 20 wdm
    FileAlternateNameInformation,   // 21
    FileStreamInformation,          // 22
    FilePipeInformation,            // 23
    FilePipeLocalInformation,       // 24
    FilePipeRemoteInformation,      // 25
    FileMailslotQueryInformation,   // 26
    FileMailslotSetInformation,     // 27
    FileCompressionInformation,     // 28
    FileObjectIdInformation,        // 29
    FileCompletionInformation,      // 30
    FileMoveClusterInformation,     // 31
    FileQuotaInformation,           // 32
    FileReparsePointInformation,    // 33
    FileNetworkOpenInformation,     // 34
    FileAttributeTagInformation,    // 35
    FileTrackingInformation,        // 36
    FileIdBothDirectoryInformation, // 37
    FileIdFullDirectoryInformation, // 38
    FileValidDataLengthInformation, // 39
    FileShortNameInformation,       // 40
    FileMaximumInformation
// begin_wdm
} FILE_INFORMATION_CLASS, *PFILE_INFORMATION_CLASS;
#else
#define FileStandardInformation ((FILE_INFORMATION_CLASS)5)
#endif

typedef struct _FILE_STANDARD_INFORMATION {                 
    LARGE_INTEGER AllocationSize;                           
    LARGE_INTEGER EndOfFile;                                
    ULONG NumberOfLinks;                                    
    BOOLEAN DeletePending;                                  
    BOOLEAN Directory;                                      
} FILE_STANDARD_INFORMATION, *PFILE_STANDARD_INFORMATION;   


bool CFileNT_Init ()
{
HMODULE h_ntdll;

//	h_psapi = LoadLibrary (TEXT("PSAPI"));
	h_ntdll = GetModuleHandle (TEXT("NTDLL"));

	if (! h_ntdll)
		return 0;

	*(FARPROC*)&ZwOpenFile = GetProcAddress (h_ntdll, "ZwOpenFile");
	*(FARPROC*)&ZwReadFile = GetProcAddress (h_ntdll, "ZwReadFile");
	*(FARPROC*)&ZwClose = GetProcAddress (h_ntdll, "ZwClose");
	*(FARPROC*)&ZwQueryInformationFile = GetProcAddress (h_ntdll, "ZwQueryInformationFile");

	return 1;
}

NTSTATUS CFileNT::Open (UNICODE_STRING* file_name)
{
	NTSTATUS status;
	NTSTATUS status_query;
	IO_STATUS_BLOCK io_status_block;
	OBJECT_ATTRIBUTES object_attributes;

	InitializeObjectAttributes(&object_attributes, file_name, 0, 0, 0);
//	NtCreateFile ();
// it cannot operate with DOS names
	status = ZwOpenFile (&handle_file, FILE_READ_ACCESS, &object_attributes, &io_status_block, FILE_SHARE_WRITE | FILE_SHARE_READ, FILE_NON_DIRECTORY_FILE);
	if ((status & 0xc0000000) == 0xc0000000)
		return status;
// seems no easy way to find out how much bytes were read, need to calculate

	FILE_STANDARD_INFORMATION file_standard_information;
	status_query = ZwQueryInformationFile (handle_file, &io_status_block, &file_standard_information, sizeof (file_standard_information), FileStandardInformation);
	file_size = file_standard_information.EndOfFile;

	return status;
}

NTSTATUS CFileNT::Open (void* file_name, int name_size)
{
	NTSTATUS status;
	UNICODE_STRING string_file_name;

	string_file_name.Buffer = (PWSTR)file_name;
	string_file_name.Length = name_size;
	string_file_name.MaximumLength = name_size;

	status = Open (&string_file_name);

	return status;
}

NTSTATUS CFileNT::Open (void* file_name)
{
	NTSTATUS status;
	int name_size;

	name_size = wcslen ((wchar_t*)file_name) * sizeof(wchar_t);

	status = Open (file_name, name_size);

	gets_offset = 0;

	return status;
}

NTSTATUS CFileNT::Read (uint64 offset, void* buffer, uint length, uint64* bytes_read)
{
	NTSTATUS status;
	IO_STATUS_BLOCK io_status_block;
	LARGE_INTEGER byte_offset;

	byte_offset.QuadPart = offset;
//	KsReadFile ();
repeat_try:
	status = ZwReadFile (handle_file, 0, 0, 0, &io_status_block, buffer, length, &byte_offset, 0);
	if (status == STATUS_PENDING)
	{
// drop time chunk
		Sleep (0);
// retry
		goto repeat_try;
	}
	if (bytes_read)
	{
		__int64 can_be_read = file_size.QuadPart - offset;

		if (length > can_be_read)
			*bytes_read = can_be_read;
		else
			*bytes_read = length;
	}

	return status;
}

NTSTATUS CFileNT::GetS (void* buffer, uint length, bool* end_of_file)
{
	NTSTATUS status;
//	bool result;
	uint64 bytes_read;
	uint i;
	uint j;
	byte symbol;
	byte* buffer_out = (byte*)buffer;

// hard limit atm
	if (length > FILENT_GETS_TYPICAL_STRING_SIZE)
		return STATUS_UNSUCCESSFUL;

// not returns eof while some bytes can been read
	status = Read (gets_offset, gets_buffer, sizeof(gets_buffer), &bytes_read);
	if (status == STATUS_END_OF_FILE)
// eof
		*end_of_file = 1;
	if ((status & 0xc0000000) >= 0x40000000)
		return status;

// dest index
	i = 0;
// src index
	j = 0;
// string ends with '\0', '\r', '\n' or eof
	while (1)
	{
//		if (i >= limit_size)
		if (i >= length)
			return STATUS_BUFFER_OVERFLOW;
		if (j >= bytes_read)
// eof
			break;
		symbol = gets_buffer[j];
		if (symbol == '\0' || symbol == '\r' || symbol == '\n')
// string end
		{
			if (symbol == '\r')
			{
				j++;
				if (j >= bytes_read)
					break;
				symbol = gets_buffer[j];
			}
			if (symbol == '\n')
				j++;
			break;
		}
		buffer_out[i] = symbol;
		i++;
		j++;
	}

	buffer_out[i] = '\0';
	gets_offset += j;
//	return STATUS_SUCCESS;
	return status;
}

bool CFileNT::Close ()
{
	NTSTATUS status;

	status = ZwClose (handle_file);
	handle_file = 0;

	return 1;
}

