// class is by Alexander Ilyin
// closest names are :  CHotKeyCtrl
// *original belongs at the Drag Watcher project
// *keys "z, x, c, v, b" would be better to leave for Winamp
// *and "up, down" better to leave for Sound Volume Hotkeys

#include "hot_keys.h"
#include "error.h"


void CHotKeys::Init ()
{
//	int i;

//	keys_num = 0;
	m_hwnd = 0;
//	for (i = 0; i < HOTKEYS_MAX; i++)
//		m_hotkeys[i].used = 0;
	memset (m_hotkeys, 0, sizeof(m_hotkeys));
}

CHotKeys::CHotKeys ()
{
	Init ();
}

CHotKeys::CHotKeys (HWND hwnd)
{
	Init ();
	SetHwnd (hwnd);
}

CHotKeys::~CHotKeys ()
{
	int i;

	for (i = 0; i < HOTKEYS_MAX; i++)
		if (m_hotkeys[i].used)
			Remove (i);
}

void CHotKeys::SetHwnd (HWND hwnd)
{
	m_hwnd = hwnd;
}

// find by modifier+key
bool CHotKeys::Find (hotkey_info_t* hotkey_info, uint* position)
{
	int i;
//	bool result;
	hotkey_t* hotkey;

	Assert(hotkey_info->function || hotkey_info->bool_ptr);

	for (i = 0; i < HOTKEYS_MAX; i++)
	{
		hotkey = &m_hotkeys[i];
		if (! m_hotkeys[i].used)
			continue;
		if (hotkey->info.modifier != hotkey_info->modifier)
			continue;
		if (hotkey->info.vkey_code != hotkey_info->vkey_code)
			continue;
		if (position)
			*position = i;
		return 1;
	}
	return 0;
}

bool CHotKeys::Add (hotkey_info_t* hotkey_info, uint* position, bool activate)
{
	int i;
	bool result;
	hotkey_t* hotkey;

	Assert(hotkey_info->function || hotkey_info->bool_ptr);

	result = Find (hotkey_info);
	if (result)
		Error ("'CHotKeys::Add ()' :  hotkey '%c' (code %02x) already is in list", hotkey_info->vkey_code, hotkey_info->vkey_code);
	for (i = 0; i < HOTKEYS_MAX; i++)
	{
		hotkey = &m_hotkeys[i];
		if (! m_hotkeys[i].used)
			break;
	}

	if (i >= HOTKEYS_MAX)
		Error ("'CHotKeys::Add ()' :  'HOTKEYS_MAX' reached");

	if (position)
		*position = i;
	hotkey->used = 1;
	hotkey->active = 0;
	hotkey->info = *hotkey_info;

	if (activate)
	{
		result = Activation (i);
		if (! result)
			Error ("'CHotKeys::Add ()' :  Activation () fail");
	}

	return 1;
}

bool CHotKeys::Remove (uint position)
{
	hotkey_t* hotkey = &m_hotkeys[position];

	Assert(position < HOTKEYS_MAX);
	if (! hotkey->used)
		return 0;
	if (hotkey->active)
		Activation (position, 0);
	hotkey->used = 0;

	return 1;
}

void CHotKeys::ActivationFailReport (uint index)
{
	hotkey_t* hotkey;
	uint modifier;
	uint vkey_code;

	hotkey = &m_hotkeys[index];
	vkey_code = hotkey->info.vkey_code;
	modifier = hotkey->info.modifier;
	m_reports_buffer[0] = 0;
	if (modifier & MOD_CONTROL)
		strcat (m_reports_buffer, "Ctrl+");
	if (modifier & MOD_SHIFT)
		strcat (m_reports_buffer, "Shift+");
	if (modifier & MOD_WIN)
		strcat (m_reports_buffer, "Win+");
	if (modifier & MOD_ALT)
		strcat (m_reports_buffer, "Alt+");
	Error ("'CHotKeys::Activation ()' :  RegisterHotKey (%02x) %s'%c' failed", vkey_code, m_reports_buffer, vkey_code);
}

bool CHotKeys::Activation (uint index, bool on)
{
BOOL B_result;
hotkey_t* hotkey = &m_hotkeys[index];

	Assert(m_hwnd);
	Assert(index < HOTKEYS_MAX);
	if (! hotkey->used)
		Error ("'CHotKeys::Activation ()' :  slot isn't used");

	if (hotkey->active == on)
		return 0;

	if (on)
// *thread-specific
	{
		B_result = RegisterHotKey (m_hwnd, index, hotkey->info.modifier, hotkey->info.vkey_code);
		if (! B_result)
			ActivationFailReport (index);
	}
	else
		B_result = UnregisterHotKey (m_hwnd, index);

	return !! B_result;
}

bool CHotKeys::ActivateAll (bool on)
{
	bool result;
	int i;
	hotkey_t* hotkey;

	result = 1;
	for (i = 0; i < HOTKEYS_MAX; i++)
	{
		hotkey = &m_hotkeys[i];
		if (hotkey->used)
			if (hotkey->active != on)
				result &= Activation (i, on);
	}
	return result;
}

void CHotKeys::OnHotKey (uint index)
{
hotkey_t* hotkey = &m_hotkeys[index];

	Assert(index < HOTKEYS_MAX);
	if (hotkey->info.function)
		hotkey->info.function (hotkey->info.user_data);
	if (hotkey->info.bool_ptr)
		*hotkey->info.bool_ptr = 1;
}

