// by Alexander Ilyin
// information about image; hits inside image collection
// *old name "region.cpp" was too much abstract
// *original belongs at the Dragging Watcher project

#include "image.h"
#include "error.h"
#include "log_file.h"
#include "displaying.h"
#include "string.h"
#include "reg_drag.h"
#include "XMessageBox.h"
#include "handle_file.h"
#include <imagehlp.h>
//#include <ntsecapi.h>


// have to redeclare 'IMAGE_DEBUG_INFORMATION', since MSVC 7.1, is aborted necessary members normal presence
// if do not use those memeber, then how this tool must work (?)
typedef struct _IMAGE_DEBUG_INFORMATION2
{
    LIST_ENTRY List;
    DWORD Size;
    PVOID MappedBase;
    USHORT Machine;
    USHORT Characteristics;
    DWORD CheckSum;
    DWORD ImageBase;
    DWORD SizeOfImage;

    DWORD NumberOfSections;
    PIMAGE_SECTION_HEADER Sections;

    DWORD ExportedNamesSize;
    PSTR ExportedNames;

    DWORD NumberOfFunctionTableEntries;
    PIMAGE_FUNCTION_ENTRY FunctionTableEntries;
    DWORD LowestFunctionStartingAddress;
    DWORD HighestFunctionEndingAddress;

    DWORD NumberOfFpoTableEntries;
    PFPO_DATA FpoTableEntries;

    DWORD SizeOfCoffSymbols;
    PIMAGE_COFF_SYMBOLS_HEADER CoffSymbols;

    DWORD SizeOfCodeViewSymbols;
    PVOID CodeViewSymbols;

    PSTR ImageFilePath;
    PSTR ImageFileName;
    PSTR DebugFilePath;

    DWORD TimeDateStamp;

    BOOL  RomImage;
    PIMAGE_DEBUG_DIRECTORY DebugDirectory;
    DWORD NumberOfDebugDirectories;

    DWORD OriginalFunctionTableBaseAddress;

// now "reserved" really :)
	DWORD Reserved[ 2 ];
} IMAGE_DEBUG_INFORMATION2, *PIMAGE_DEBUG_INFORMATION2;

static uint64 image__total_registering_memory = 0;


CWatchImage::CWatchImage ()
{
	have_file_path = 0;
	*file_path = 0;
	*file_name = 0;
	main_exe = 0;
	time_stamp = 0;
	image_base = 0;
	image_size = 0;
	registration_array_size = 0;
	registration_array = 0;
//	c_map_file = 0;
	map_file_loaded = 0;
	hits_counter = 0;
	displaying_dump_boolean_mask = 0;
}

CWatchImage::~CWatchImage ()
{
}


// +++ initialization +++

void CWatchImage::ZeroMem ()
{
	Assert(image_size >= 0);
	ZeroMemory(registration_array, registration_array_size);
}

bool CWatchImage::Init (bool is_main_exe, HANDLE file_handle, void* image_base_address)
{
	bool result;
//	MEMORY_BASIC_INFORMATION memory_basic_information;
//	char buffer[UPATH_MAX];
	IMAGE_DEBUG_INFORMATION2* image_debug_information2;
//	DWORD D_result;
	BOOL B_result;
	char* exe_type_name;
	UNICODE_STRING* unicode_string = (UNICODE_STRING*)file_path;
	uint i;

	main_exe = is_main_exe;
	exe_type_name = main_exe ? "exe" : "dll";
//	base_of_image = image_base_address;
//	have_file_path = GetFileNameByHandle (file_handle, buffer, UPATH_MAX);
//	have_file_path = HandleFile_GetFileName (process_id, h_process, base_of_image, file_handle, (UNICODE_STRING*)buffer, UPATH_MAX);
	have_file_path = HandleFile_GetFileName (0, regdrag_h_process, image_base_address, file_handle, unicode_string, UPATH_MAX);
	if (have_file_path)
		LogFile_Printf ("%s filename is \"%S\"\n", exe_type_name, unicode_string->Buffer);
	else
	{
		LogFile_Printf ("%s filename is not given\n", exe_type_name);
		file_path[0] = 0;
	}

	image_debug_information2 = (IMAGE_DEBUG_INFORMATION2*)MapDebugInformation (file_handle, 0, 0, 0);

	if (! image_debug_information2)
		LogFile_Printf ("----- Can't get \"image debug information\" -----\n");
	else
	{
// no no no ! this address is often invalid, 10000000 (it is adress in the file, as is easy to suppose)
//		image_base = pimage_debug_information2->ImageBase;
		image_base = (uint)image_base_address;
// *but this size looks proper : it isn't just size of a file, and appears is size of that image
		image_size = image_debug_information2->SizeOfImage;
		image_end = image_base + image_size;
		sections_num = image_debug_information2->NumberOfSections;
		if (sections_num > SECTIONS_MAX)
			Error ("CWatchImage::Init () :  the number of sections (%d) is too high (%d is max).", sections_num, SECTIONS_MAX);
		result = c_map_file.Init (unicode_string->Buffer, file_name, sections, sections_num);
		for (i = 0; i < sections_num; i++)
		{
// store section info
			::Section_Init (&sections[i], &image_debug_information2->Sections[i]);
// inform the mapfile member about the size of section
			c_map_file.StoreFunction (i, sections[i].size, MAPFILE_CTYPE__NONE, "--- end of section ---", "--- invalid ---");
		}
		time_stamp = image_debug_information2->TimeDateStamp;
//		LogFile_Printf ("\"time stamp\" = %08x\n", time_stamp);
		B_result = UnmapDebugInformation ((IMAGE_DEBUG_INFORMATION*)image_debug_information2);
	}

#if 0
// get the size
	D_result = VirtualQueryEx (h_process, section_base, &memory_basic_information, sizeof(MEMORY_BASIC_INFORMATION));

	section_size = memory_basic_information.RegionSize;
	LogFile_Printf ("%s section size is %08x\n", exe_type_name, section_size);
#endif

	Assert(image_size >= 0);
	registration_array_size = sizeof(*registration_array) * image_size;
	registration_array = (uint*)malloc (registration_array_size);
	ZeroMem ();
	image__total_registering_memory += registration_array_size;
//	LogFile_Printf ("Image \"%S\", base %08x, size %08x\n", (have_file_path ? unicode_string->Buffer : L"unknown"), image_base, image_size);
	LogFile_Printf ("Image base %08x, size %08x, time stamp %08x\n", image_base, image_size, time_stamp);

	if (main_exe ? read_map_main_exe : read_map_loaded_dlls)
	{
		result = ReadMapFile ();
		return result;
	}

	return 1;
}

void CWatchImage::Quit ()
{
	Assert(!! registration_array);
	free (registration_array);
	registration_array = 0;
	registration_array_size = 0;
}

// returns 0, if user refuse to continue, in case of no .map file found
// so, 0 doesn't means "not loaded"
bool CWatchImage::ReadMapFile (void)
{
//bool result;
//int map_file_loaded = 0;
bool map_file_abort = 0;
//char file_name[256];
//char map_file_path[128];
char* c_result;
//ushort* u_result;
//char* path;
bool abort = 0;
//BOOL B_result;
wchar_t* wc_ptr;
UNICODE_STRING* unicode_string = (UNICODE_STRING*)file_path;

	if (! have_file_path)
// don't know the name, to make a try
		return 1;

//	c_map_file = new CMapFile;

// read process memory will not be necessary, all info is in 'IMAGE_DEBUG_INFORMATION'
//	B_result = ReadProcessMemory (h_process, base_of_image, displaying_block, section_size, 0);

// any map file must be near it's own executable file; otherwise, becomes hard to manage what is where
//	path = compile_info_path ? compile_info_path : regdrag_dir_process_executable;
//	map_file_loaded = c_map_file->LoadFile (map_file_name, path, time_stamp);
// extract module name
	wc_ptr = wcsrchr (unicode_string->Buffer, L'\\');
	if (! wc_ptr)
		Error ("CWatchImage::ReadMapFile () :  path without '\\'");
	wc_ptr++;
	sprintf (file_name, "%.255S", wc_ptr);
	file_name[255] = 0;
// find extension
	c_result = strrchr (file_name, '.');
	if (c_result)
// remove extension
		*c_result = 0;
	map_file_loaded = c_map_file.LoadFile (time_stamp, &map_file_abort);

#if 0
	if (map_file_loaded != 1)
// map file isn't loaded
	{
	int mb_result;
	char text[256];
	static bool skip_all_not_found = 0;

//		LogFile_Printf ("----- Map file \"%s\\%s.map\" is not loaded. -----\n", path, map_file_name);

		delete c_map_file;
		c_map_file = 0;

		if (! map_file_loaded)
		{
			if (! skip_all_not_found)
			{
				sprintf (text, "Map file \"%s\" is not loaded from directory :  \"%s\"\nContinue anyway ?", map_file_name, path);

//				mb_result = MessageBox (0,text,0,MB_YESNO);
				mb_result = XMessageBox (0, text, 0, MB_YESNO | MB_YESTOALL, 0);

				if (mb_result == IDNO)
					abort = 1;
				if (mb_result == IDYESTOALL)
					skip_all_not_found = 1;
			}
		}
		else
// abort requested
			abort = 1;
	}
	else
		LogFile_Printf ("Map file \"%s\\%s.map\" has been loaded.\n", path, map_file_name);

	return ! abort;
#endif

	return ! map_file_abort;
}

// --- initialization ---


bool CWatchImage::IsAddressInImage (uint address)
{
	if (address < image_base)
		return 0;
	if (address >= image_end)
		return 0;
	return 1;
}

// register the break at this address
// *return value isn't used at this moment
bool CWatchImage::Hit (uint address)
{
DWORD cur_address_offset;

	hits_counter++;

//	GetOffset (address, &cur_address_offset);
	cur_address_offset = address - image_base;

	if (! registration_array)
		Error ("'registration_array' isn't initialized");

	if (registration_array[cur_address_offset] < 0xffffffff)
	{
		registration_array[cur_address_offset]++;
		if (registration_array[cur_address_offset] == 0xffffffff)
			LogFile_Printf (file_name ? "--- Counter limit is hit in %s ---\n" : "--- Counter limit is hit ---\n", file_name);
	}

	return 0;
}

#if 0
void CWatchImage::ShowFunction (uint address)
{
//DWORD offset;

// todo
#if 0
	GetOffset (address, &offset);

// *it is not a request of higher class
	Displaying_ShowFunction (this, (ulong)offset);
#endif
}
#endif

void CWatchImage::Exception (uint address)
{
	IMAGE_DEBUG_INFORMATION2* pimage_debug_information2;

	pimage_debug_information2 = (IMAGE_DEBUG_INFORMATION2*)MapDebugInformation (0, 0, 0, address);
	if (! pimage_debug_information2)
		LogFile_Printf ("----- CWatchImage::Exception () :  Can't get \"image debug information\" for getting \"module name\" -----\n");
	else
	{
		LogFile_Printf ("module name is \"%s\"\n\n", pimage_debug_information2->DebugFilePath);
		UnmapDebugInformation ((IMAGE_DEBUG_INFORMATION*)pimage_debug_information2);
	}
//	ShowFunction (address);
}

bool CWatchImage::FindSection (int* section_index, uint offset)
{
	uint i;

	for (i = 0; i < sections_num; i++)
	{
		section_t* section = &sections[i];
		if (offset < section->offset_begin)
			continue;
		if (offset >= section->offset_end)
			continue;
		*section_index = i;
		return 1;
	}

	return 0;
}

bool CWatchImage::GetFunctionInfo (map_file_function_info_t** map_file_function_info, uint offset)
{
	int section_index;
	bool result;
	uint offset_in_section;
	section_t* section;

	Assert(!! map_file_function_info);

	if (! map_file_loaded)
		return 0;

	result = FindSection (&section_index, offset);
	if (! result)
		return 0;

	section = &sections[section_index];
	offset_in_section = offset - section->offset_begin;

	result = c_map_file.Section_GetFunctionInfo (map_file_function_info, section_index, offset);
	if (! result)
		return 0;

#if 0
// append more info
	map_file_function_info_t* func_info;
	func_info = (*map_file_function_info);
	func_info->section_index = section_index;
	func_info->offset_in_image = section->offset_begin + func_info->offset_begin;
#endif

	return 1;
}

void Image_ShowTotalMemoryAllocated ()
{
	uint mb;

// round it to MB
	mb = image__total_registering_memory / (1024 * 1024 / 2);
	mb++;
	mb /= 2;
	LogFile_Printf ("Total memory allocated for registering :  %08I64x bytes (%u MB)\n\n", image__total_registering_memory, mb);
}

