// by Alexander Ilyin
// sorting of got debug results

//#include<assert.h>

#include "types.h"
#include "sorting.h"
#include "reg_drag.h"
#include "error.h"
#include "log_file.h"


// --- settings ---

// - blocks -
// how close placed points to unite 
// *distance is to one side
uint distance_unite = 0x30;

// --- end of settings ---

module_info_t modules[IMAGES_MAX];
point_info_t* points = NULL;
int num_points = 0;
block_info_t* blocks = NULL;
int num_blocks = 0;
function_info_t* functions = NULL;
int num_functions = 0;


// +++ by modules +++

void Sorting_Module_Put (module_info_t* module_info)
{
bool move;
module_info_t val0, val1;
module_info_t* module;
int i;

	for (move = 0, i = 0; i < IMAGES_MAX; i++)
	{
		module = &modules[i];

		if (! move)
// moving, currently isn't activated
		{
			if (module->image_index < 0 || module_info->hits_num > module->hits_num)
// from here - counters not exist or are lower
			{
// actvate moving part of points list forward, begginning from current
				move = 1;
				val0 = *module_info;
			}
		}

// *note :  not 'else'
		if (move)
// continue compare with counters isn't needed - do moving
		{
// remember lower value
			val1 = *module;
// put newest
			*module = val0;
// prepare for the next (if will be)
			val0 = val1;
		}
	}
}

void Sorting_ByModules ()
{
int img;

	for (img = 0; img < images_num; img++)
	{
	module_info_t module_info;

		module_info.image_index = img;
		module_info.hits_num = c_images[img]->hits_counter;

		Sorting_Module_Put (&module_info);
	}
}

// --- by modules ---


// +++ by points +++

void Sorting_Point_Put (point_info_t* point_info)
{
int i;
bool move;
point_info_t val0, val1;

	for (move = 0, i = 0; i < num_points; i++)
	{
		point_info_t* point = &points[i];

		if (! move)
// moving, currently isn't activated
		{
			if (point->image_index < 0 || point->hits_number < point_info->hits_number)
// from here - counters not exist or are lower
			{
// actvate moving part of points list forward, begginning from current
				move = 1;
				val0 = *point_info;
			}
		}

// *note :  not 'else'
		if (move)
// continue compare with counters isn't needed - do moving
		{
// remember lower value
			val1 = *point;
// put newest
			*point = val0;
			val0 = val1;
		}
	}
}

// find out mostly dragging points
void Sorting_ByPoints_Image (int img)
{
uint i;
uint* registration_array;
SIZE_T image_size;
point_info_t* point_last = &points[num_points - 1];
point_info_t point_info;

	point_info.image_index = img;
	registration_array = c_images[img]->registration_array;
	image_size = c_images[img]->image_size;

	for (i = 0; i < image_size; i++)
	{
		uint point_hits_number = registration_array[i];

// speed up
		if (! point_hits_number)
// nothing was registered here
			continue;
// speed up
// compare with last element in the table
		if (point_last->image_index >= 0 && point_hits_number <= point_last->hits_number)
			continue;

		point_info.address = i;
		point_info.hits_number = point_hits_number;

		Sorting_Point_Put (&point_info);
	}
}

void Sorting_ByPoints ()
{
int img;

	for (img = 0; img < images_num; img++)
		Sorting_ByPoints_Image (img);
}

// --- by points ---


// +++ by blocks +++

void Sorting_Block_Put (block_info_t* block_info)
{
bool move;
block_info_t val0, val1;
block_info_t* block;
int i;

	for (move = 0, i = 0; i < num_blocks; i++)
	{
		block = &blocks[i];

		if (! move)
// moving, currently isn't activated
		{
			if (block->image_index < 0 || block_info->number > block->number)
// from here - counters not exist or are lower
			{
// actvate moving part of points list forward, begginning from current
				move = 1;
				val0 = *block_info;
			}
		}

// *note :  not 'else'
		if (move)
// continue compare with counters isn't needed - do moving
		{
// remember lower value
			val1 = *block;
// put newest
			*block = val0;
// prepare for the next (if will be)
			val0 = val1;
		}
	}
}

// find out mostly dragging blocks
void Sorting_ByBlocks_Image (int img)
{
uint i;
block_info_t block_info;
uint* registration_array;
SIZE_T image_size;

	block_info.image_index = img;
	block_info.start = -1;
	registration_array = c_images[img]->registration_array;
	image_size = c_images[img]->image_size;

	for (i = 0; i < image_size; i++)
	{
		uint point_hits_number = registration_array[i];

// skip - only for registered
		if (! point_hits_number)
			continue;

		if (block_info.start == -1)
// at the start of first block
		{
			block_info.start = i;
			block_info.mostly = i;
			block_info.end = i;
			block_info.number = point_hits_number;
			block_info.number_for_mostly = point_hits_number;
			continue;
		}
		else
// ordinary continue
		{
			if (i - block_info.end > distance_unite)
// next block
			{
				Sorting_Block_Put (&block_info);
				block_info.start = i;
				block_info.mostly = i;
				block_info.end = i;
				block_info.number = point_hits_number;
				block_info.number_for_mostly = point_hits_number;
			}
			else
// still will be this block
			{
				block_info.number += point_hits_number;

				if (point_hits_number > block_info.number_for_mostly)
// set this as mostly dragging point in the block
				{
// index
					block_info.mostly = i;
// number
					block_info.number_for_mostly = point_hits_number;
				}

				block_info.end = i;
			}
		}
	}

// put last block
	if (block_info.start != -1)
		Sorting_Block_Put (&block_info);
}

void Sorting_ByBlocks ()
{
int img;

	for (img = 0; img < images_num; img++)
		Sorting_ByBlocks_Image (img);
}

// --- by blocks ---


// +++ by functions +++

void Sorting_Function_Put (function_info_t* function_info)
{
bool move;
function_info_t val0, val1;
function_info_t* function;
int i;

	for (move = 0, i = 0; i < num_functions; i++)
	{
		function = &functions[i];

		if (! move)
// moving, currently isn't activated
		{
//			if (function->image_index < 0 || function_info->number > function->number)
			if (! function->number || function_info->number > function->number)
// from here - counters not exist or are lower
			{
// actvate moving part of points list forward, begginning from current
				move = 1;
// write at once (prepare) new info
				val0 = *function_info;
			}
		}

// *note :  not 'else'
		if (move)
// continue compare with counters isn't needed - do moving
		{
// remember lower value
			val1 = *function;
// put newest
			*function = val0;
// prepare for the next (if will be)
			val0 = val1;
		}
	}
}

// find out mostly dragging functions
void Sorting_ByFunctions_Section (CWatchImage* c_image, int sect, section_t* section)
{
//CWatchImage* c_image = c_images[img];
uint i;
function_info_t function_info;
uint* registration_array;
//uint section_base = c_image->image_base + section->offset_begin;
uint section_offset = section->offset_begin;
uint section_size = section->size;
map_file_function_info_t* map_file_function_infos;
int number_of_functions;
map_file_section_info_t* map_file_section_info;

	map_file_section_info = &c_image->c_map_file.section_infos[sect];
//	map_file_section_info = &c_image->c_map_file;
//	function_info.image_index = img;
	function_info.c_image = c_image;
// force index defining
	function_info.function_index = -1;
	function_info.map_file_function_info = 0;
	registration_array = c_image->registration_array;
//	image_size = c_image->image_size;
	map_file_function_infos = map_file_section_info->map_file_function_infos;
	number_of_functions = map_file_section_info->number_of_functions;
// last entry is the end marker
	number_of_functions--;

	if (section_offset > c_image->image_size)
	{
		LogFile_Printf ("Sorting :  module \"%s\", section %d offset %08x is beyond image end %08x\n", c_image->file_name, sect, section_offset, c_image->image_size);
		return;
	}
	if (section_offset + section_size > c_image->image_size)
	{
		LogFile_Printf ("Sorting :  module \"%s\", section %d end %08x is beyond image end %08x\n", c_image->file_name, sect, section_offset + section_size, c_image->image_size);
		return;
	}

	for (i = 0; i < section_size; i++)
	{
		uint point_hits_number = registration_array[section_offset + i];

// skip; only for registered
		if (! point_hits_number)
			continue;

		if (function_info.function_index >= number_of_functions)
		{
			Error ("Sorting_ByFunctions_Section () :  module \"%s\", section %d, functions list (%d entries) unexpectedly finished\n", c_image->file_name, sect, number_of_functions);
		}

// if (not this function)
		if (i >= map_file_function_infos[function_info.function_index + 1].offset_begin)
// another function
		{
			if (function_info.function_index != -1)
				Sorting_Function_Put (&function_info);

// search for function
			for (function_info.function_index++; ; function_info.function_index++)
			{
				if (function_info.function_index >= number_of_functions)
					Error ("Sorting_ByFunctions_Section () :  module \"%s\", section %d, functions list (%d entries), no entry for offset %08x\n", c_image->file_name, sect, number_of_functions, i);
				if (i < map_file_function_infos[function_info.function_index].offset_begin)
				{
// it is previous function
					function_info.function_index--;
					break;
				}
			}

// -- init new function info --
// *not just '0'
			function_info.map_file_function_info = &map_file_function_infos[function_info.function_index];
			function_info.offset_mostly = i - map_file_function_infos[function_info.function_index].offset_begin;
			function_info.number = point_hits_number;
			function_info.number_for_mostly = point_hits_number;
		}
		else
// still will be this function
		{
			function_info.number += point_hits_number;

			if (point_hits_number > function_info.number_for_mostly)
// set this as mostly dragging point in the function
			{
				Assert(i >= map_file_function_infos[function_info.function_index].offset_begin);
// offset
				function_info.offset_mostly = i - map_file_function_infos[function_info.function_index].offset_begin;
// number
				function_info.number_for_mostly = point_hits_number;
			}
		}
	}

// put last function
	if (function_info.function_index != -1)
		Sorting_Function_Put (&function_info);
}

void Sorting_ByFunctions_Image (int img)
{
uint i;
CWatchImage* c_image = c_images[img];

	if (! c_image->map_file_loaded)
// map file isn't loaded
		return;

	for (i = 0; i < c_image->sections_num; i++)
	{
		section_t* section = &c_image->sections[i];
		if (section->is_executable)
			Sorting_ByFunctions_Section (c_image, i, section);
	}
}

void Sorting_ByFunctions ()
{
int img;

	for (img = 0; img < images_num; img++)
		Sorting_ByFunctions_Image (img);
}

// --- by functions ---
