// by Alexander Ilyin
// *original belongs at the Dragging Watcher project

#include "string.h"
#include <string.h>
#include <stdarg.h>
#include "error.h"
#if OWN_VSNPRINTF
#include "vsnprintf.h"
#endif


uint Strlen (const char* str)
{
	uint len = 0;
	while (*str)
	{
		len++;
		str++;
	}
	return len;
}

// 'strncpy ()' copy at most n characters of string -> remaked -> 'Strncpy ()' copy into buffer of size n
// return 1, if string was copied with it's '\0'
bool Strncpy (char* dest, const char* src, uint buffer_size)
{
// *leave one char for '\0'
	if (buffer_size < sizeof(*dest))
// no buffer
		return 0;

	while (1)
	{
// copy string character
		if (! (*dest = *src))
// string successfully finished
			return 1;

// and only now check that the buffer have no more place
		if (buffer_size == sizeof(*dest))
// no more place in buffer
		{
// finish string
			*dest = L'\0';
			return 0;
		}

		dest++;
		src++;
		buffer_size--;
	}

//	return 0;
}

bool Strcpy (char* dest, const char* src)
{
	bool result;
	result = Strncpy (dest, src, -1);
	return result;
}

// 'wcsncpy ()'
// return 1, if string was copied with it's L'\0'
bool WStrncpy (wchar_t* dest, const wchar_t* src, uint buffer_size)
{
// lets simplificate calculations below
	buffer_size &= 0xfffffffe;

// *leave one char for '\0'
	if (buffer_size < sizeof(*dest))
// no buffer
		return 0;

	while (1)
	{
// copy string character
		if (! (*dest = *src))
// string successfully finished
			return 1;

		if (buffer_size == sizeof(*dest))
// no more place in buffer
		{
// finish string
			*dest = L'\0';
			return 0;
		}

		dest++;
		src++;
		buffer_size--;
	}

//	return 0;
}

bool Strncat (char* dest, char* src, uint buffer_size)
{
	bool result;
	uint buffer_remain;
	uint len;

	len = Strlen (dest);
// but allow to try to append a string (empty for example), if the buffer already is full (not overflown)
	if (len >= buffer_size)
		return 0;
	buffer_remain = buffer_size - len;
	result = Strncpy (dest + len, src, buffer_remain);
	return result;
}

bool Strcat (char* dest, char* src)
{
	bool result;
	result = Strncat (dest, src, -1);
	return result;
}

// copy, and return 0, if it is not alphanum or does not fit into buffer
bool Str_IsAlphaNum (const char* str)
{
//	bool result;
	char c;

//	result = Strncpy (dest, src, buffer_size);
//	if (! result)
//		return 0;
	while (c = *str)
	{
		if (c >= 'a' && c <= 'z')
			goto continue_loop;
		if (c >= 'A' && c <= 'Z')
			goto continue_loop;
		if (c >= '0' && c <= '9')
			goto continue_loop;
// this one allow also
		if (c == '_')
			goto continue_loop;
// not alphanum character
		return 0;
continue_loop:
		str++;
	}
// string successfully finished
	return 1;
}

// insert one char in a string in sized buffer
bool StringN_InsertChar (char* string, uint string_buffer_size, char char_insert, uint position)
{
uint string_full_lenght = Strlen (string) + 1;
uint i;

	if (string_full_lenght >= string_buffer_size)
// no place to insert
		return 0;

	i = string_full_lenght;

// copy the string one symbol forward
	while (i-- > position)
//		string[i] = string[--i];
		string[i + 1] = string[i];

// place the symbol
	string[position] = char_insert;

	return 1;
}

// delete one char, to the left from the position
bool String_BackspaceChar (char* string, uint position)
{
uint string_lenght = Strlen (string);

	if (! position || position > string_lenght)
// no char, to the left from the position
		return 0;

// copy the string one symbol back
	while (string[position - 1] = string[position])
		position++;

	return 1;
}

// delete one char, on the position
bool String_DeleteChar (char* string, uint position)
{
uint string_lenght = Strlen (string);

	if (position >= string_lenght)
// no char, on the position
		return 0;

// copy the string one symbol back
	while (string[position] = string[position + 1])
		position++;

	return 1;
}

bool String_UnicodeToANSI (char* ansi_string, int buffer_size, ushort* unicode_string, int string_size)
{
int i;
int chars_number = string_size / sizeof(*unicode_string);
bool result = 1;

	if (chars_number >= buffer_size)
	{
		chars_number = buffer_size - 1;
		result = 0;
	}

	for (i = 0; i < chars_number; i++)
		ansi_string[i] = (char)unicode_string[i];

	ansi_string[i] = '\0';

	return result;
}


#if OWN_VSNPRINTF
// lets make simplier subroutines simplier to type; although 'Vsnprintf ()' and 'Vsprintf ()' are not in this rule :)
// 'Snprintf ()' and 'Sprintf ()' simpliest;
// 'SNPrintf ()' and 'SPrintf ()' with strlen returned;
// 'SNPrintF ()' and 'SPrintF ()' (all words begins with upper case letter) are 'snprintf ()' and 'sprintf ()' clones, differing that they takes not symbols printed max value, but buffer size;

// am not sure what to return through that bool result :/

//	_snprintf (string, buffer_size - 1, format, ...);
bool SNPrintf (char* buffer, uint buffer_size, uint* symbols_written, const char* format, ...)
{
	bool result;
	va_list argptr;

	va_start (argptr, format);
	result = ! Vsnprintf (buffer, buffer_size, format, argptr, symbols_written);
	va_end (argptr);
	return result;
}

bool Snprintf (char* buffer, uint buffer_size, const char* format, ...)
{
	bool result;
	va_list argptr;

	va_start (argptr, format);
	result = ! Vsnprintf (buffer, buffer_size, format, argptr);
	va_end (argptr);
	return result;
}

// 'snprintf ()' clone; use a define to point at this function, do not rename this function
int SNPrintF (char* buffer, uint buffer_size, const char* format, ...)
{
	bool result;
	uint symbols_written;
	va_list argptr;

	va_start (argptr, format);
	result = ! Vsnprintf (buffer, buffer_size, format, argptr, &symbols_written);
	va_end (argptr);
	if (! result)
		return -1;
	return symbols_written;
}

bool Vsprintf (char* buffer, uint* symbols_written, const char* format, va_list argptr)
{
	bool result;

	result = ! Vsnprintf (buffer, -1, format, argptr, symbols_written);
	return result;
}

bool SPrintf (char* buffer, uint* symbols_written, const char* format, ...)
{
	bool result;
	va_list argptr;

	va_start (argptr, format);
	result = Vsprintf (buffer, symbols_written, format, argptr);
	va_end (argptr);
	return result;
}

bool Sprintf (char* buffer, const char* format, ...)
{
	bool result;
	va_list argptr;

	va_start (argptr, format);
	result = Vsprintf (buffer, 0, format, argptr);
	va_end (argptr);
	return result;
}

// 'sprintf ()' clone; use a define to point at this function, do not rename this function
int SPrintF (char* buffer, const char* format, ...)
{
	bool result;
	uint symbols_written;
	va_list argptr;

	va_start (argptr, format);
	result = Vsprintf (buffer, &symbols_written, format, argptr);
	va_end (argptr);
	if (! result)
		return -1;
	return symbols_written;
}
#endif
